/*
 * Copyright (c) 1997, 1998, 1999
 *	Alex Feldman <al.feldman@sangoma.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Alex Feldman.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Alex Feldman AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Alex Feldman OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: sdla_fr.c,v 1.14 2004/12/28 14:31:32 sangoma Exp $
 */

/*****************************************************************************
 * sdla_fr.c	WANPIPE(tm) Multiprotocol WAN Link Driver. Frame relay module.
 *
 * Author(s):	Alex Feldman  <al.feldman@sangoma.com>
 *
 * ===========================================================================
 * Sep 17, 2003 Alex Feldman		o Fixed ALTQ bug (IFQ_ENQUEUE)
 * Nov 30, 2001 Alex Feldman		o Added T1/E1 support (TE1).
 * Mar 10, 2000 Alex Feldman RX_CORR	o If receive buffer corrupted in 
 *					  rx_intr(), disable interrupt!!! 
 * Dec 10, 1999 Alex Feldman		o Updated to FreeBSD 3.2
 * Jun 02, 1999	Gideon Hack		o Updated for S514 support.
 * Jan 02, 1997	Gene Kozin		o Initial version.
 ****************************************************************************
 */
/*
****************************************************************************
****		For Debug purpose (only OpenBSD)			****
****************************************************************************
*/
#ifdef __OpenBSD__
# undef WANDEBUG_FR	/* Uncomment this line for debug output */
# undef ENABLE_IPV6	/* Uncomment this line for Enabling IPv6 support */
#endif

#include <net/wanpipe_includes.h>
#define _GNUC_
#include <net/wanpipe_snmp.h>
#include <net/wanpipe.h>
#include <net/if_wanpipe_common.h>
#include <net/sdla_fr.h>

/****** Defines & Macros ****************************************************/

#define	MAX_CMD_RETRY	10		/* max number of firmware retries */

#define	FR_HEADER_LEN	8		/* max encapsulation header size */
#define	FR_CHANNEL_MTU	1500		/* unfragmented logical channel MTU */

/* IP over Frame Relay encapsulation ( 2 bytes )*/
#define FR_IP_HEADER			0xCC03

/* InArp request retransmission timeout in seconds */
#define INARP_REQUEST_TIMEOUT 20

/* Q.922 frame types */
#define	Q922_UI		0x03		/* Unnumbered Info frame */
#define	Q922_XID	0xAF		

/* CISCO data frame encapsulation */
#define CISCO_UI	0x08
#define CISCO_IP	0x00
#define CISCO_INV	0x20

/* DLCI configured or not */
#define DLCI_NOT_CONFIGURED	0x00
#define DLCI_CONFIG_PENDING	0x01
#define DLCI_CONFIGURED		0x02

/* CIR enabled or not */
#define CIR_ENABLED	0x00
#define CIR_DISABLED	0x01

#define WANPIPE 0x00
#define API     0x01
#define FRAME_RELAY_API 1

#define FR_MBOX_OFFS		0x0
#define FR_TX_MBOX_OFFS		0x800

/* For handle_IPXWAN() */
#define CVHexToAscii(b) (((unsigned char)(b) > (unsigned char)9) ? ((unsigned char)'A' + ((unsigned char)(b) - (unsigned char)10)) : ((unsigned char)'0' + (unsigned char)(b)))

#define read_eflags()	({register u_long ef;		\
		__asm("pushfl; popl %0" : "=r" (ef));	\
		ef;})
#define write_eflags(x)	({register u_long ef = (x);	\
		__asm("pushl %0; popfl" : : "r" (ef));})

/****** Data Structures *****************************************************/

/* This is an extention of the 'struct device' we create for each network
 * interface to keep the rest of channel-specific data.
 */
typedef struct fr_channel
{
	wanpipe_common_t	common;
	char name[WAN_IFNAME_SZ+1];	/* interface name, ASCIIZ */
	unsigned dlci_configured  ;	/* check whether configured or not */
	unsigned cir_status;		/* check whether CIR enabled or not */
	unsigned dlci;			/* logical channel number */
	unsigned cir;			/* committed information rate */
	unsigned bc;			/* committed burst size */
	unsigned be;			/* excess burst size */
	unsigned mc;			/* multicast support on or off */
	unsigned tx_int_status;		/* Transmit Interrupt Status */	
	unsigned short pkt_length;	/* Packet Length */
	unsigned long router_start_time;/* Router start time in seconds */
	int tick_counter;	/* counter for transmit time out */
	char dev_pending_devtint;	/* interface pending dev_tint() */
	unsigned int dlci_int_interface_off; /* offset to the DLCI Interface */ 
	unsigned long IB_addr;		/* physical address of Interface Byte */
	unsigned long state_tick;	/* time of the last state change */
	unsigned char enable_IPX;	/* Enable/Disable the use of IPX */
	unsigned long network_number;	/* Internal Network Number for IPX*/
	unsigned route_flag;		/* Add/Rem dest addr in route tables */
	unsigned inarp;			/* Inverse Arp Request status */ 
	int inarp_interval;		/* Time between InArp Requests */
	unsigned long inarp_tick;	/* InArp jiffies tick counter */
	if_send_stat_t drvstats_if_send;
        rx_intr_stat_t drvstats_rx_intr;
        pipe_mgmt_stat_t drvstats_gen;

	unsigned char usedby;  /* Used by WANPIPE or API */

	unsigned long router_up_time;

	unsigned short transmit_length;
	char transmit_buffer[FR_MAX_NO_DATA_BYTES_IN_FRAME];

        /* Special members for FreeBSD */
  	int         start;     /* 1/0 - opened/closed interface (dev->start) */ 
        int         fr_inarp_reply_expected; /* */
        int	    fr_inarp_reply_timer;    /* timer counter for reply */
        int         fr_resend_inarp;	     /* set when we need to resend*/
        arphdr_fr_t fr_inarp_pkt;	     /* this packet 		     */
	/* new destination address */
	unsigned long               dst_addr;

	/* SNMP */
	unsigned long 		router_last_change;
	unsigned int		dlci_type;
	unsigned long		rx_DE_set;
	unsigned long		tx_DE_set;
	unsigned long		rx_FECN;
	unsigned long		rx_BECN;
	unsigned 		err_type;
	char			err_data[SNMP_FR_ERRDATA_LEN];
	unsigned long 		err_time;
	unsigned long 		err_faults;
	unsigned		trap_state;
	unsigned long		trap_max_rate;

	unsigned char		fr_encap_0;
	unsigned char		fr_encap_1;

	unsigned short		timer_int_enabled;
} fr_channel_t;

/* Route Flag options */
#define NO_ROUTE	0x00
#define ADD_ROUTE 	0x01
#define ROUTE_ADDED	0x02
#define REMOVE_ROUTE 	0x03

/* inarp options */
#define INARP_NONE		0x00
#define INARP_REQUEST		0x01
#define INARP_CONFIGURED	0x02

/* reasons for enabling the timer interrupt on the adapter */
#define TMR_INT_ENABLED_UDP   		0x0001
#define TMR_INT_ENABLED_UPDATE 		0x0002
#define TMR_INT_ENABLED_INARP 		0x0004
#define TMR_INT_ENABLED_TE 		0x0008
#define TMR_INT_ENABLED_UPDATE_DLCI	0x0010
#define TMR_INT_ENABLED_TEST		0x0020
#define TMR_INT_ENABLED_PROCESS_ROUTE	0x0040

typedef struct dlci_status
{
	unsigned short dlci	PACKED;
	unsigned char state	PACKED;
} dlci_status_t;

typedef struct dlci_IB_mapping
{
	unsigned short dlci		PACKED;
	unsigned long  addr_value	PACKED;
} dlci_IB_mapping_t;

/* This structure is used for DLCI list Tx interrupt mode.  It is used to
   enable interrupt bit and set the packet length for transmission
 */
typedef struct fr_dlci_interface 
{
	unsigned char gen_interrupt	PACKED;
	unsigned short packet_length	PACKED;
	unsigned char reserved		PACKED;
} fr_dlci_interface_t;

/* variable for keeping track of enabling/disabling FT1 monitor status */
static int rCount = 0;

#ifdef __OpenBSD__
extern int ticks;
#endif

extern void disable_irq(unsigned int);
extern void enable_irq(unsigned int);

/* variable for keeping track of number of interrupts generated during 
 * interrupt test routine 
 */
static int Intr_test_counter;
/****** Function Prototypes *************************************************/

/* WAN link driver entry points. These are called by the WAN router module. */
static int update(wan_device_t *wandev);
static int new_if(wan_device_t *wandev, struct ifnet *ifp, wanif_conf_t *conf);
static int del_if(wan_device_t *wandev, struct ifnet *ifp);

/* WANPIPE-specific entry points */
static int wpf_exec(struct sdla *card, void *u_cmd, void *u_data);

/* Network device interface */
static int fr_close(struct ifnet *ifp);

/* ifnet entry points */
static int fr_output   		/* output routine (enqueue) */
		__P((struct ifnet *, struct mbuf *, struct sockaddr *,
		      struct rtentry*));		
static void fr_start 		/* initiate output routine  */
		__P((struct ifnet *));
static int fr_ioctl		/* ioctl routine */
		__P((struct ifnet *, u_long, caddr_t));
static void fr_watchdog		/* timer routine */
		__P((struct ifnet*));
static int fr_init		/* init routine */
		__P((struct ifnet*));

static void fr_up(struct ifnet *);
static void fr_down(struct ifnet *);

static int chk_bcast_mcast_addr(sdla_t *card, struct ifnet* ifp,
                                struct mbuf *m);

/* Interrupt handlers */
static void fr_isr(sdla_t* card);
static void rx_intr(sdla_t *card);
static void tx_intr(sdla_t *card);
static void timer_intr(sdla_t *card);
static void spur_intr(sdla_t *card);

/* Frame relay firmware interface functions */
static int fr_read_version(sdla_t *card, char *str);
static int fr_configure(sdla_t *card, fr_conf_t *conf);
static int fr_dlci_configure(sdla_t *card, fr_dlc_conf_t *conf, unsigned dlci);
static int fr_init_dlci(sdla_t *card, fr_channel_t *chan);
static int fr_set_intr_mode (sdla_t *card, unsigned mode, unsigned mtu, unsigned short timeout);
static int fr_comm_enable(sdla_t *card);
static int fr_comm_disable(sdla_t *card);
static int fr_get_err_stats(sdla_t *card, fr_comm_stat_t* stats);
static int fr_get_stats(sdla_t *card, int dlci, void* stats);
static int fr_get_dlci_stats(sdla_t *card, fr_channel_t* chan);
static int fr_add_dlci(sdla_t *card, int dlci);
static int fr_activate_dlci(sdla_t *card, int dlci);
static int fr_delete_dlci (sdla_t* card, int dlci);
static int fr_issue_isf(sdla_t *card, int isf);
static int fr_write(sdla_t *card, int dlci, unsigned char attr, int len);
static int fr_send(sdla_t *card, int dlci, unsigned char attr, int len,
	void *buf, int snd_encp);

/* Firmware asynchronous event handlers */
static int fr_event(sdla_t *card, int event, wan_mbox_t *mbox);
static int fr_modem_failure(sdla_t *card, wan_mbox_t *mbox);
static int fr_dlci_change(sdla_t *card, wan_mbox_t *mbox);

/* Miscellaneous functions */
static int update_chan_state(struct ifnet *ifp);
static void set_chan_state(struct ifnet *ifp, int state);
static struct ifnet *find_channel(sdla_t *card, unsigned dlci);
static int is_tx_ready(sdla_t *card, fr_channel_t *chan);
static unsigned int dec_to_uint(unsigned char *str, int len);
static int reply_udp(wan_udp_pkt_t* udp_pkt, unsigned int mbox_len ); 

#ifdef FR_OLD_INTR_TEST
static int fr_intr_test(sdla_t* card);
#endif
static void init_chan_statistics( fr_channel_t* chan );
static void init_global_statistics( sdla_t* card );
static void read_DLCI_IB_mapping( sdla_t* card, fr_channel_t* chan );
static void setup_for_delayed_transmit(struct ifnet* ifp);
static void process_route (netdevice_t *ifp);


/* Inverse ARP and Dynamic routing functions */
int process_ARP(arphdr_1490_t *ArpPacket, sdla_t *card, struct ifnet *ifp);
int is_arp(void *buf);
static void send_inarp_request(sdla_t *card, struct ifnet *ifp);
static void send_inarp_packet(sdla_t *card, arphdr_fr_t *arphdr, 
						struct ifnet *ifp);

/* Udp management functions */
static int process_udp_mgmt_pkt(sdla_t *card);
static int udp_pkt_type(sdla_t* card, struct mbuf *m);
static int store_udp_mgmt_pkt(int udp_type, char udp_pkt_src, sdla_t* card,
                                struct mbuf *m, int dlci);

/* IPX functions */
#ifdef SUPPORT_IPX
static void switch_net_numbers(unsigned char *sendpacket,
			unsigned long network_number, unsigned char incoming);
static int handle_IPXWAN(unsigned char *sendpacket, char *devname,
			unsigned char enable_IPX, unsigned long network_number);
#endif

extern struct mbuf* sdla_get(struct ifnet* ifp, char*, unsigned);

static WRITE_FRONT_END_REG_T write_front_end_reg;
static READ_FRONT_END_REG_T  read_front_end_reg;
static void fr_enable_timer(void* card_id);
static int set_adapter_config (sdla_t* card);
static void fr_handle_front_end_state (void* card_id);

/* SNMP */
static int fr_snmp_data(sdla_t* card, struct ifnet* ifp, void* data);

static int fr_udp_ioctl(struct ifnet* ifp, caddr_t data);
static int process_udp_cmd(struct ifnet* ifp, wan_udp_hdr_t* udp_hdr);

static int fr_debugging(sdla_t* card);
static unsigned long fr_crc_frames(sdla_t* card);
static unsigned long fr_abort_frames(sdla_t * card);
static unsigned long fr_tx_underun_frames(sdla_t* card);

/* Procfs functions */
static int wan_fr_get_info(void* pcard, struct seq_file* m, int* stop_cnt); 

/****** Public Functions ****************************************************/

/*============================================================================
 * Frame relay protocol initialization routine.
 *
 * This routine is called by the main WANPIPE module during setup.  At this
 * point adapter is completely initialized and firmware is running.
 *  o read firmware version (to make sure it's alive)
 *  o configure adapter
 *  o initialize protocol-specific fields of the adapter data space.
 *
 * Return:	0	o.k.
 *		< 0	failure.
 */
int wpf_init(sdla_t *card, wandev_conf_t *conf)
{
	int err;
	union{
		char str[80];
        	fr_conf_t cfg;
	} u;
	fr_buf_info_t buf_info;
	int i;
	wan_smp_flag_t	s;

	wan_spin_lock_irq(NULL, &s);

	DEBUG_INIT("%s: Starting Frame Relay initialization\n", 
						card->wandev.name);
	/* Verify configuration ID */
	if (conf->config_id != WANCONFIG_FR){ 
		DEBUG_EVENT("%s: invalid configuration ID %u!\n",
				card->devname, conf->config_id);
		wan_spin_unlock_irq(NULL, &s);
		return -EINVAL;
	}

	/* Initialize protocol-specific fields of adapter data space */
	card->mbox_off = FR_MB_VECTOR + FR508_MBOX_OFFS; 
       	card->flags_off = FR_MB_VECTOR + FR508_FLAG_OFFS;
	card->intr_type_off = 
			card->flags_off +
			offsetof(fr508_flags_t, iflag);
	card->intr_perm_off = 
			card->flags_off +
			offsetof(fr508_flags_t, imask);

	card->isr = &fr_isr;

	err = card->hw_iface.check_mismatch(card->hw,conf->fe_cfg.media);
	if (err){
		wan_spin_unlock_irq(NULL, &s);
		return err;
	}

	/* Determine the board type user is trying to configure
	 * and make sure that the HW matches */
	if (IS_TE1_MEDIA(conf->fe_cfg.media)) {

		memcpy(&card->fe.fe_cfg, &conf->fe_cfg, sizeof(sdla_fe_cfg_t));
		card->fe.name		= card->devname;
		card->fe.card		= card;
		card->fe.write_fe_reg	= write_front_end_reg;
		card->fe.read_fe_reg	= read_front_end_reg;

		card->wandev.te_enable_timer = fr_enable_timer;
		card->wandev.te_link_state = fr_handle_front_end_state;
		conf->interface = 
			(IS_T1_CARD(card)) ? WANOPT_V35 : WANOPT_RS232;
		conf->clocking = WANOPT_EXTERNAL;
		
        }else if (IS_56K_MEDIA(conf->fe_cfg.media)) {
                
		memcpy(&card->fe.fe_cfg, &conf->fe_cfg, sizeof(sdla_fe_cfg_t));
		card->fe.name		= card->devname;
		card->fe.card		= card;
		card->fe.write_fe_reg	= write_front_end_reg;
		card->fe.read_fe_reg	= read_front_end_reg;

	}else{
		card->fe.fe_status = FE_CONNECTED;
	}

	/* Read firmware version.  Note that when adapter initializes, it
	 * clears the mailbox, so it may appear that the first command was
	 * executed successfully when in fact it was merely erased. To work
	 * around this, we execute the first command twice.
	 */
	if (fr_read_version(card, NULL) || fr_read_version(card, u.str)){
		wan_spin_unlock_irq(NULL, &s);
		return -EIO;
	}

	DEBUG_EVENT("%s: running frame relay firmware v%s\n", 
			card->devname, u.str);

	if (set_adapter_config(card)){
		wan_spin_unlock_irq(NULL, &s);
		return -EIO;
	}

	/* Adjust configuration */
	conf->mtu = (conf->mtu >= MIN_LGTH_FR_DATA_CFG) ?
			min(conf->mtu, FR_MAX_NO_DATA_BYTES_IN_FRAME) :
                        FR_CHANNEL_MTU;
     
	conf->bps = min(conf->bps, 2048000);

	/* Initialze the configuration structure sent to the board to zero */
	bzero(&u.cfg, sizeof(u.cfg));
	bzero(&card->u.f, sizeof(sdla_fr_t));

	bzero(card->u.f.dlci_to_dev_map, sizeof(card->u.f.dlci_to_dev_map));

	/* Configure adapter firmware */
	u.cfg.mtu	= conf->mtu + FR_HEADER_LEN + 10; /* Add another 10 to be sure */
	u.cfg.kbps	= conf->bps / 1000;

	u.cfg.cir_fwd = u.cfg.cir_bwd = 16;
	u.cfg.bc_fwd  = u.cfg.bc_bwd = 16;

	u.cfg.options	= 0x0000;
	DEBUG_EVENT("%s: Global CIR enabled by Default\n", card->devname);

	switch (conf->u.fr.signalling){
	case WANOPT_FR_ANSI:
        	u.cfg.options = 0x0000; 
		break;		

	case WANOPT_FR_Q933:	
		u.cfg.options |= 0x0200; 
		break;

	case WANOPT_FR_LMI:	
		u.cfg.options |= 0x0400; 
		break;

	case WANOPT_NO:
		u.cfg.options |= 0x0800; 
		break;

	default:
		DEBUG_EVENT("%s: Illegal Signalling option\n", 
					card->wandev.name);
		wan_spin_unlock_irq(NULL, &s);
		return -EINVAL;
	}
	card->wandev.signalling = conf->u.fr.signalling;

	if (conf->u.fr.station == WANOPT_CPE){
		if (conf->u.fr.signalling == WANOPT_NO){
			DEBUG_EVENT("%s: For NO signalling, station must be set to Node!\n",
							card->devname);
			wan_spin_unlock_irq(NULL, &s);
			return -EINVAL; 
		}
		u.cfg.station = 0;
		u.cfg.options |= FRCFG_AUTO_CONFIG;	/* auto config DLCI */
		card->u.f.dlci_num  = 0;
	}else{ 
		u.cfg.station = 1;	/* switch emulation mode */

		/* For switch emulation we have to create a list of dlci(s)
		 * that will be sent to be global SET_DLCI_CONFIGURATION 
		 * command in fr_configure() routine. 
		 */
		card->u.f.dlci_num  = min(max(conf->u.fr.dlci_num, 1), 100);

		for ( i = 0; i < card->u.f.dlci_num; i++){ 
			card->u.f.node_dlci[i] = (unsigned short) 
		              conf->u.fr.dlci[i] ? conf->u.fr.dlci[i] : 16;
		}
	}

	if (conf->clocking == WANOPT_INTERNAL)
		u.cfg.port |= 0x0001;

	if (conf->interface == WANOPT_RS232)
		u.cfg.port |= 0x0002;

	if (conf->u.fr.t391)
		u.cfg.t391 = min(conf->u.fr.t391, 30);
	else
		u.cfg.t391 = 5;
	card->u.f.t391 = u.cfg.t391;

	if (conf->u.fr.t392)
		u.cfg.t392 = min(conf->u.fr.t392, 30);
	else
		u.cfg.t392 = 15;
	card->u.f.t392 = u.cfg.t392;

	if (conf->u.fr.n391)
		u.cfg.n391 = min(conf->u.fr.n391, 255);
	else
		u.cfg.n391 = 2;
	card->u.f.n391 = u.cfg.n391;

	if (conf->u.fr.n392)
		u.cfg.n392 = min(conf->u.fr.n392, 10);
	else
		u.cfg.n392 = 3;	
	card->u.f.n392 = u.cfg.n392;

	if (conf->u.fr.n393)
		u.cfg.n393 = min(conf->u.fr.n393, 10);
	else
		u.cfg.n393 = 4;
	card->u.f.n393 = u.cfg.n393;

	if (fr_configure(card, &u.cfg)){
		wan_spin_unlock_irq(NULL, &s);
		return -EIO;
	}

	card->hw_iface.peek(card->hw, 
			    FR_MB_VECTOR + FR508_RXBC_OFFS, 
			    &buf_info, 
			    sizeof(fr_buf_info_t));
	card->rxmb_off = buf_info.rse_next; 
	card->u.f.rxmb_base_off = buf_info.rse_base;
	card->u.f.rxmb_last_off = 
			buf_info.rse_base + (buf_info.rse_num - 1) * sizeof(fr_rx_buf_ctl_t);

	card->u.f.rx_base_off = buf_info.buf_base;
	card->u.f.rx_top_off  = buf_info.buf_top;

	card->u.f.tx_interrupts_pending = 0;

	card->wandev.mtu	   = conf->mtu;
	card->wandev.bps	   = conf->bps;
	card->wandev.interface = conf->interface;
	card->wandev.clocking  = conf->clocking;
	card->wandev.station   = conf->u.fr.station;
	card->poll		   = NULL; 
	card->exec		   = &wpf_exec;
	card->wan_debugging	   = &fr_debugging;
	card->get_crc_frames   = &fr_crc_frames;
	card->get_abort_frames = &fr_abort_frames;
	card->get_tx_underun_frames	= &fr_tx_underun_frames;
	card->wandev.update	   = &update;
	card->wandev.new_if	   = &new_if;
	card->wandev.del_if	   = &del_if;
	card->wandev.ttl	   = conf->ttl;
	card->wandev.udp_port  = conf->udp_port; 
	/* Update card state */
	card->wandev.state	   = WAN_DISCONNECTED;

	/* SNMP data */
	card->get_snmp_data = &fr_snmp_data;

	card->wandev.get_info		= &wan_fr_get_info;

	/* Intialize global statistics for a card */
	init_global_statistics( card );

	card->TracingEnabled          = 0;

	/* Interrupt Test */
	wan_spin_unlock_irq(NULL, &s);

#ifdef FR_OLD_INTR_TEST
	Intr_test_counter = 0;
	card->intr_mode = INTR_TEST_MODE;
	err = fr_intr_test( card );

	if (err || (Intr_test_counter < MAX_INTR_TEST_COUNTER)){ 
		DEBUG_EVENT("%s: Interrupt Test Failed, Counter: %d\n", 
					card->devname, Intr_test_counter);
		DEBUG_EVENT("Please choose another interrupt\n");
		return -EIO;
	}

	DEBUG_EVENT("%s: Interrupt Test Passed, Counter: %d\n",
					card->devname, Intr_test_counter);
#endif

	/* Configure the front end.
	 * If the onboard CSU/DSU exists, configure via
	 * user defined options */
	if (IS_TE1_CARD(card)){

		DEBUG_EVENT("%s: Configuring onboard %s CSU/DSU\n",
			card->devname, 
			(IS_T1_CARD(card))?"T1":"E1");
		if (sdla_te_config(&card->fe, &card->wandev.fe_iface)){
	
			DEBUG_EVENT("%s: Failed %s configuratoin!\n",
				card->devname,
			 	(IS_T1_CARD(card))?"T1":"E1");
			return -EIO;
		}
	
	}else if (IS_56K_CARD(card)){

		DEBUG_EVENT("%s: Configuring 56K onboard CSU/DSU\n",
				card->devname);

		if(sdla_56k_config(&card->fe, &card->wandev.fe_iface)){
			DEBUG_EVENT("%s: Failed 56K configuration!\n",
				card->devname);
			return -EIO;
		}
	}

	return 0;
}

/******* WAN Device Driver Entry Points *************************************/

/*============================================================================
 * Update device status & statistics.
 */
static int update (wan_device_t* wandev)
{
	sdla_t          *card = NULL;
	unsigned long timeout;

	/* sanity checks */
	if ((wandev == NULL) || (wandev->private == NULL))
		return -EFAULT;

	if (wandev->state == WAN_UNCONFIGURED)
		return -ENODEV;

	card = wandev->private;

	card->update_comms_stats = 1;
	/* SNMP */
	card->u.f.timer_int_enabled |= 
		(card->u.f.update_dlci == NULL) ? TMR_INT_ENABLED_UPDATE:
					          TMR_INT_ENABLED_UPDATE_DLCI;
	card->hw_iface.set_bit(card->hw, card->intr_perm_off, FR_INTR_TIMER);
	timeout = SYSTEM_TICKS;
	for(;;){
		if (card->update_comms_stats == 0)
			break;
		if ((SYSTEM_TICKS - timeout) > (1 * hz)){
			card->update_comms_stats = 0;
			return -EAGAIN;
		}
	}
	return 0;
}

/*============================================================================
 * Create new logical channel.
 * This routine is called by the router when ROUTER_IFNEW IOCTL is being
 * handled.
 * o parse media- and hardware-specific configuration
 * o make sure that a new channel can be created
 * o allocate resources, if necessary
 * o prepare network device structure for registaration.
 *
 * Return:	0	o.k.
 *		< 0	failure (channel will not be created)
 */
static int new_if (wan_device_t* wandev, struct ifnet* ifp, wanif_conf_t* conf)
{
	sdla_t* card = wandev->private;
	fr_channel_t* chan = NULL;
	int dlci = 0;
	int err = 0;

	if ((conf->name[0] == '\0') || (strlen(conf->name) > WAN_IFNAME_SZ)){ 
		DEBUG_EVENT("%s: invalid interface name!\n", card->devname);
		return -EINVAL;
	}

	if (wandev->new_if_cnt > 0 && card->u.f.auto_dlci_cfg){
		DEBUG_EVENT("%s:%s: Error, DLCI autoconfig only allowed on a single DLCI\n",
			    card->devname,chan->name);
		DEBUG_EVENT("%s: Please turn off DLCI autoconfig in %s.conf\n",
				card->devname,card->devname);
		DEBUG_EVENT("\n");
		return -EINVAL;
	}
	
	/* verify media address */
	if (strcmp(conf->addr, "auto")){

		if (is_digit(conf->addr[0])){ 
			dlci = dec_to_uint(conf->addr, 0);
	
			if (!dlci || (dlci > HIGHEST_VALID_DLCI)){ 
				DEBUG_EVENT("%s:%s: invalid DLCI %u on interface!\n",
						card->devname, conf->name, dlci);
				return -EINVAL;
			}
		}else{ 
			DEBUG_EVENT("%s:%s: invalid media address on interface!\n",
						card->devname, conf->name);
			return -EINVAL;
		}
	}else{
		/* FIXME: Add supporting Auto config option for Frame Relay */
		DEBUG_EVENT("%s: Error, DLCI autoconfig options not supported!\n",
					card->devname);
		DEBUG_EVENT("%s: Please turn off DLCI autoconfig in %s.conf\n",
				card->devname,card->devname);
		return -EINVAL;

		if (card->wandev.station == WANOPT_NODE){
			DEBUG_EVENT("%s: Error, DLCI autoconfig only allowed for CPE!\n",
					card->devname);
			DEBUG_EVENT("%s: Please turn off DLCI autoconfig in %s.conf\n",
				card->devname,card->devname);
			return -EINVAL;
		}
		
		if (wandev->new_if_cnt != 0){
			DEBUG_EVENT("%s:%s: Error, DLCI autoconfig only allowed on a single DLCI\n",
						card->devname,chan->name);
			DEBUG_EVENT("%s: Please turn off DLCI autoconfig in %s.conf\n",
						card->devname,card->devname);
			return -EINVAL;
		}

		card->u.f.auto_dlci_cfg=1;

		dlci=0xFFFF;
	}
		
	/* allocate and initialize private data */
	chan = wan_malloc(sizeof(fr_channel_t));

	if (chan == NULL){
		return -ENOMEM;
	}

	bzero(chan, sizeof(fr_channel_t));
	chan->common.card = card;
	chan->dlci = dlci;

	/* Copy network interface name to chan */ 
	memcpy(chan->name, conf->name, strlen(conf->name));

	if (card->u.f.auto_dlci_cfg){
		DEBUG_EVENT("%s: Interface %s configured for AUTO DLCI\n",
					card->devname,conf->name);
	}else{
		DEBUG_EVENT("%s: Interface %s bound to DLCI %d\n",
					card->devname,conf->name,dlci);
	}

	/* Setup wanpipe as a router (WANPIPE) or as an API */
	if (strcmp(conf->usedby, "WANPIPE") == 0){
        	DEBUG_EVENT("%s: WANPIPE mode enabled for %s\n",
					card->devname, chan->name);
	        chan->usedby = WANPIPE;
	}else if(strcmp(conf->usedby, "API") == 0){
#ifdef FRAME_RELAY_API
		chan->usedby = API;
		DEBUG_EVENT("%s: API(%d) mode enabled for %s\n", 
				card->devname, chan->usedby, chan->name);
#else
		DEBUG_EVENT("%s: API Mode is not supported !\n", 
				card->devname);
		DEBUG_EVENT("%s: API patch can be obtained from Sangoma Tech.\n",
			     	card->devname);
		err = -EINVAL;
#endif
	}

	if (err){ 
        	wan_free(chan);
	        return err;
	}

	card->u.f.dlci_to_dev_map[dlci] = ifp;

	/* place cir,be,bc and other channel specific information into the
	** chan structure 
	*/
	if (conf->cir){ 
		chan->cir = max( 1, min( conf->cir, 512 ) );
		chan->cir_status = CIR_ENABLED; 

		/* If CIR is enabled, force BC to equal CIR
	         * this solves number of potential problems if CIR is 
        	 * set and BC is not 
	         */
        	chan->bc = chan->cir;

	        if (conf->be){
        	    chan->be = max(0, min(conf->be, 511)); 
	        }else{	
        	    conf->be = 0;
		}

	        DEBUG_EVENT("%s: CIR ENABLED for DLCI %d \n", 
				card->devname,chan->dlci);
		DEBUG_EVENT("%s:     CIR = %d ; BC = %d ; BE = %d\n",
				card->devname,chan->cir,chan->bc,chan->be);
	}else{
		chan->cir_status = CIR_DISABLED;
		DEBUG_EVENT("%s: CIR_DISABLED for DLCI %d\n", 
				card->devname,chan->dlci);
	}

	chan->mc = conf->mc;

	chan->inarp = conf->inarp ? INARP_REQUEST : INARP_NONE;

	chan->inarp_interval = 
			conf->inarp_interval ? conf->inarp_interval : 10;

	chan->dlci_configured = DLCI_NOT_CONFIGURED;

	/* FIXME: IPX disabled in this WANPIPE version */
	if (conf->enable_IPX == WANOPT_YES){
		DEBUG_EVENT("%s: This version of WANPIPE doesn't support IPX\n",
						card->devname);
		wan_free(chan);
		return -EINVAL;
	}else{
		chan->enable_IPX = WANOPT_NO;
	}
	chan->enable_IPX = conf->enable_IPX;

	if (conf->network_number){
		chan->network_number = conf->network_number;
	}else{
		chan->network_number = 0xDEADBEEF;
	}

	chan->route_flag = NO_ROUTE;

	init_chan_statistics(chan);

	chan->transmit_length = 0;

	/* prepare network device data space for registration */
	if (wan_netif_init(ifp, chan->name)){
		DEBUG_EVENT("%s:%s: Failed to initialize network interface!\n", 
					card->devname, conf->name);
		wan_free(chan);
		return -ENOMEM;
	}
	ifp->if_softc          = chan;
	ifp->if_mtu            = wandev->mtu;
	ifp->if_type           = IFT_OTHER;
	ifp->if_flags          = IFF_POINTOPOINT;
	ifp->if_output         = fr_output;
	ifp->if_start          = fr_start;
	ifp->if_watchdog       = fr_watchdog;
	ifp->if_ioctl          = fr_ioctl;
#ifdef __FreeBSD__
	ifp->if_init           = (void*)fr_init;
#endif
	ifp->if_snd.ifq_len    = 0;

	ifp->if_hdrlen = FR_HEADER_LEN;
	/* For an API, the maximum number of bytes that the stack will pass
	 * to the driver is (dev->mtu + dev->hard_header_len). So, adjust the
	 * mtu so that a frame of maximum size can be transmitted by the API. 
	 */
	if (chan->usedby == API){ 
		ifp->if_mtu += (sizeof(api_tx_hdr_t) - FR_HEADER_LEN);
	}

	set_chan_state(ifp, WAN_DISCONNECTED);

	/* Enable Interrupts and Communications */
	if (!wandev->new_if_cnt){
	
		wandev->new_if_cnt++;
		/* 
		 * If you enable comms and then set ints, you get a Tx int 
		 * as you perform the SET_INT_TRIGGERS command. So, we only 
		 * set int triggers and then adjust the interrupt mask (to 
		 * disable Tx ints) before enabling comms. 
		 */	
	        if (fr_set_intr_mode(card, (FR_INTR_RXRDY | FR_INTR_TXRDY |
			FR_INTR_DLC | FR_INTR_TIMER | FR_INTR_TX_MULT_DLCIs | 
			FR_INTR_MODEM), card->wandev.mtu, 0))
	        {
			wan_free(chan);
			return -EIO;
		}
#ifdef FR_OLD_INTR_TEST
		card->hw_iface.clear_bit(card->hw, card->intr_perm_off, (FR_INTR_TXRDY | FR_INTR_TIMER));
#else
		card->hw_iface.clear_bit(card->hw, 
					    card->intr_perm_off,
					    (FR_INTR_TXRDY | FR_INTR_RXRDY | FR_INTR_DLC |
				 FR_INTR_TX_MULT_DLCIs | FR_INTR_MODEM));
#endif
 
#ifdef FR_OLD_INTR_TEST
		if (fr_comm_enable(card)){
			wan_free(chan);
			return -EIO;
		}	
#else
		Intr_test_counter = 0;
		card->u.f.timer_int_enabled |= TMR_INT_ENABLED_TEST;
		ifp->if_timer = 5;
#endif
		/* Manually poll the 56K CSU/DSU to get the status */
		if (IS_56K_CARD(card)){
			/* 56K Update CSU/DSU alarms */
			card->wandev.fe_iface.read_alarm(&card->fe, 1);
		}

#ifdef FR_OLD_INTR_TEST
		wanpipe_set_state(card, WAN_CONNECTED);
#endif
	}

	/* Initialize the encapsulation header to the IETF standard */
	chan->fr_encap_0 = Q922_UI;
	chan->fr_encap_1 = NLPID_IP;

	/* SNMP */
	chan->dlci_type = SNMP_FR_STATIC;
	chan->trap_state = SNMP_FR_DISABLED;
	chan->trap_max_rate = 0;

	return 0;
}

/*
 * ============================================================================
 * Delete logical channel.
 */
static int del_if (wan_device_t* wandev, struct ifnet* ifp)
{
	sdla_t*		card = wandev->private;
	fr_channel_t* 	chan = (fr_channel_t*)ifp->if_softc;

	if (chan == NULL){
		return 0;
	}
	/* Execute shutdown very first time we enter del_if */
	DEBUG_EVENT("%s: Removing interface %s...\n", 
				card->devname, chan->name);
	if (!wandev->del_if_cnt){
		wandev->del_if_cnt++;
		wanpipe_set_state(card, WAN_DISCONNECTED);
		/* Initialize Interrupt handler pointer */
		card->isr = NULL;  
		fr_set_intr_mode(card, 0, 0, 0);
		fr_comm_disable(card);
		if (IS_TE1_CARD(card)){
			sdla_te_unconfig(card);
		}
	}

	/* Close current interface and remove DLCI */ 
	fr_close(ifp);

	/* Initialize newtwork interface */
	WAN_IFQ_PURGE(&ifp->if_snd);
	ifp->if_output = NULL; 
	ifp->if_start = NULL; 
	ifp->if_ioctl = NULL; 
	ifp->if_watchdog = NULL; 
#ifdef __FreeBSD__
	ifp->if_init = NULL;
#endif
	ifp->if_flags &= ~(IFF_UP | IFF_RUNNING | IFF_OACTIVE); 
	if (ifp->if_softc){
		wan_free(ifp->if_softc);
		ifp->if_softc = NULL;
	}
	wan_netif_del(ifp);
	return 0;
}

/****** WANPIPE-specific entry points ***************************************/

/*
 * ============================================================================
 * Execute adapter interface command.
 */
static int wpf_exec (struct sdla* card, void* u_cmd, void* u_data)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int 	retry = MAX_CMD_RETRY;
	int 	err, len;
	
	if (copyin(u_cmd, (void*)mbox, sizeof(wan_cmd_t))){
		return -EFAULT;
	}

	/* execute command */
	do {
		if (mbox->wan_data_len){
			if (copyin(u_data, (void*)mbox->wan_data, mbox->wan_data_len)){
				return -EFAULT;
			}	
		}
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	} while (err && retry-- && !fr_event(card, err, mbox));
	/* return result */
	if (copyout((void*)&mbox->wan_command, u_cmd, sizeof(wan_cmd_t))){
		return -EFAULT;
	}	
	len = mbox->wan_data_len;
	if (len && u_data){
	       	if (copyout((void*)mbox->wan_data, u_data, len)){
			return -EFAULT;
		}
	}
			
	return 0;
}

/****** Network Device Interface ********************************************/

/*
 * fr_ioctl() -
 *
 */
static int
fr_ioctl(struct ifnet * ifp, u_long command, caddr_t data)
{
    fr_channel_t*	chan = (fr_channel_t*)ifp->if_softc;
    sdla_t*		card = (sdla_t*)chan;
    struct ifaddr*	ifa = (struct ifaddr*)data;
    wan_smp_flag_t	s;		    	    /* saved interrupt state */
    int		      	error = 0;	    	    /* resulting status */

    wan_spin_lock_irq(NULL, &s);

    switch (command){ 
    case SIOCSIFADDR:
        DEBUG_IOCTL("%s:%s: Ioctl (SIOCSIFADDR) called\n",
				card->devname, chan->name);
        /* check the protocol family - we currently support IP only */
        switch(ifa->ifa_addr->sa_family){ 
        case AF_INET:
	    if (fr_init(ifp)){
		return EBUSY; 
	    }
            break;
        default:
            DEBUG_EVENT("%s:%s: Unsupported address family (%d)\n",
				card->devname, chan->name, ifa->ifa_addr->sa_family);
            error = EAFNOSUPPORT;
            break;
        }
        break;

    case SIOCGIFADDR:
        DEBUG_IOCTL("%s:%s: Ioctl (SIOCGIFADDR) called.\n", 
				card->devname, chan->name);
        break;

    case SIOCSIFFLAGS:
        DEBUG_IOCTL("%s:%s: Ioctl (SIOCSIFFLAGS) called.\n", 
				card->devname, chan->name);
        break;

    case SIOCGIFFLAGS:
        DEBUG_IOCTL("%s:%s: Ioctl (SIOCGIFFLAGS) called.\n", 
				card->devname, chan->name);
        break;

    case SIOCSIFMTU:
        DEBUG_IOCTL("%s:%s: Ioctl (SIOCSIFMTU) called.\n", 
				card->devname, chan->name);
        error = EINVAL;
        break;

    case SIOC_WANPIPE_PIPEMON:
  	DEBUG_IOCTL("%s:%s: Ioctl (SIOC_WANPIPE_PIPEMON) called.\n",
        			card->devname, chan->name);
	error = fr_udp_ioctl(ifp, data);
	break;

    case SIOC_WANPIPE_SNMP:
    case SIOC_WANPIPE_SNMP_IFSPEED:
	wan_spin_unlock_irq(NULL, &s);
	return wan_snmp_data(card, ifp, command, (struct ifreq*)data);
	break;

    default:
        DEBUG_IOCTL("%s:%s: Ioctl (default) called.\n", 
				card->devname, chan->name);
        error = EINVAL;
        break;
    }

    wan_spin_unlock_irq(NULL, &s);
    return error;
}

/*
 * ============================================================================
 * fr_up() - bring interface UP and RUNNING
 */
static void fr_up(struct ifnet *ifp)
{
	ifp->if_flags |= (IFF_UP | IFF_RUNNING);
}

/*
 *============================================================================
 * fr_down() - Bring interface not UP and NOT RUNNING
 */
static void fr_down(struct ifnet *ifp)
{
	ifp->if_flags &= ~(IFF_UP | IFF_RUNNING);
}

/*============================================================================
 * if_wachdog():
 *
 * 
 * 
 * 
 */
static void fr_watchdog(struct ifnet * ifp)
{
	fr_channel_t *chan = ifp->if_softc;   /* our softc struct */
	sdla_t       *card = chan->common.card;

	/* see if we've perviously failed to send an InARP packet  - if so */
	/* resend it and reset the flag ...				   */
	if (chan->fr_resend_inarp){ 
		/* clear the flag */
		chan->fr_resend_inarp = 0;
		/* resend */
		DEBUG_TIMER("%s:%s: Watchdog tries to send InARP packet ...\n",
					card->devname, wan_netif_name(ifp));
		send_inarp_packet(chan->common.card, &chan->fr_inarp_pkt, ifp);
	}

	/* see if we're awaiting for InARP reply - if so decrement the timer */
	/* and if it goes off - resend the request			     */
	if (chan->fr_inarp_reply_expected){ 
		if (--chan->fr_inarp_reply_timer == 0){ 
			send_inarp_request(chan->common.card, ifp);
		}else{ 
			/* make sure the watchdog is called again in a second */
			ifp->if_timer = 1;
		}
	}

	if (chan->timer_int_enabled & TMR_INT_ENABLED_PROCESS_ROUTE){
		process_route(ifp);
		chan->timer_int_enabled &= ~TMR_INT_ENABLED_PROCESS_ROUTE;
	}

#ifndef FR_OLD_INTR_TEST   
	if (card->u.f.timer_int_enabled & TMR_INT_ENABLED_TEST){ 
    		if (Intr_test_counter < MAX_NEW_INTR_TEST_COUNTER){ 
			card->u.f.timer_int_enabled &= 
					~TMR_INT_ENABLED_TEST; 
			DEBUG_EVENT("%s: Interrupt Test Failed, Counter: %d\n", 
					card->devname, Intr_test_counter);
			DEBUG_EVENT("Please choose another interrupt\n");
			fr_set_intr_mode(card, 0, 0, 0);
    		}
	}
#endif
}

/* Ok
 * ============================================================================
 * Initialize network interface.
 *
 * This routine is called only once for each interface, during Linux network
 * interface registration.  Returning anything but zero will fail interface
 * registration.
 */
static int fr_init (struct ifnet* ifp)
{
	fr_channel_t*	chan = ifp->if_softc;
	sdla_t*		card = chan->common.card;
	wan_smp_flag_t	s;
	int			err = 0;

	if (chan->start){
		return -EBUSY;	/* only one open is allowed */
	}
  
    	wan_spin_lock_irq(NULL, &s);
	if (card->wandev.signalling == WANOPT_NO){
		DEBUG_INIT("%s: Signalling set to NO: Mapping DLCI's\n", 
					card->devname);
		if (fr_init_dlci(card, chan)){
			wan_spin_unlock_irq(NULL, &s);
			return -EAGAIN;
		}
	}

	if (card->wandev.station == WANOPT_CPE){ 
		/* CPE: issue full status enquiry */
		fr_issue_isf(card, FR_ISF_FSE);
	}else{ /* FR switch: activate DLCI(s) */
		
		/* For Switch emulation we have to ADD and ACTIVATE
		 * the DLCI(s) that were configured with the SET_DLCI_
		 * CONFIGURATION command. Add and Activate will fail if
		 * DLCI specified is not included in the list.
		 *
		 * Also If_open is called once for each interface. But
		 * it does not get in here for all the interface. So
 		 * we have to pass the entire list of DLCI(s) to add 
		 * activate routines.  
		 */ 

		fr_add_dlci(card, chan->dlci);
		fr_activate_dlci(card, chan->dlci);
	}

	DEBUG_EVENT("%s: Initialization successful for DLCI %d\n", 
					card->devname, chan->dlci);

	chan->start = 1; /* identical to dev->start in Linux */
	wanpipe_open(card);
	update_chan_state(ifp);
  	wan_getcurrenttime(&chan->router_start_time, NULL);
	wan_spin_unlock_irq(NULL, &s);
	return err;
}

/* 
 * ============================================================================
 * Close network interface.
 * o if this is the last open, then disable communications and interrupts.
 * o reset flags.
 */
static int fr_close (struct ifnet* ifp)
{
	fr_channel_t* 	chan = ifp->if_softc;
	sdla_t*       	card = chan->common.card;
	wan_smp_flag_t	s;

	wan_spin_lock_irq(NULL, &s);

	chan->start = 0; /* identical to dev->start in Linux */
	wanpipe_close(card);
	if (card->wandev.station == WANOPT_NODE){
		fr_delete_dlci (card,chan->dlci);
	}

	DEBUG_INIT("%s: IF CLOSE: Successful for DLCI %d\n",
			card->devname, chan->dlci);

    	wan_spin_unlock_irq(NULL, &s);
	return 0;
}

/* 
 * ============================================================================
 * Send a packet on a network interface.
 * o set tbusy flag (marks start of the transmission) to block a timer-based
 *   transmit from overlapping.
 * o set critical flag when accessing board.
 * o check link state. If link is not up, then drop the packet.
 * o check channel status. If it's down then initiate a call.
 * o pass a packet to corresponding WAN device.
 * o free socket buffer
 *
 * Return:	0	complete (socket buffer must be freed)
 *		non-0	packet may be re-transmitted (tbusy must be set)
 *
 * Notes:
 * 1. This routine is called either by the protocol stack or by the "net
 *    bottom half" (with interrupts enabled).
 * 2. Setting tbusy flag will inhibit further transmit requests from the
 *    protocol stack and can be used for flow control with protocol layer.
 */
static int
fr_output(struct ifnet *ifp, struct mbuf *m, struct sockaddr *dst, struct rtentry *rt0)
{
	fr_channel_t*	chan = ifp->if_softc;
	sdla_t*		card = chan->common.card;
	unsigned char*	sendpacket = NULL;
	void*		data = NULL;
	int             udp_type = 0, 
			send_data = 0;
	unsigned int    len = 0;
	wan_smp_flag_t	s;		       	/* saved int. state */
	int 		error = 0;
#if defined(ALTQ)
	struct altq_pktattr	pktattr;
#endif

	chan->drvstats_if_send.if_send_entry ++;
	if (m == NULL){ 
		/* if we get here, some higher layer thinks we've missed an
		 * tx-done interrupt.
		 */
		DEBUG_EVENT("%s: interface %s got kicked!\n", 
					card->devname,chan->name);
		chan->drvstats_if_send.if_send_skb_null ++;
		return 0;
	}

	if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING)){
		wan_skb_free(m);	/*m_freem(m);*/
		return ENETDOWN;
	}

	if ((card->wandev.state == WAN_CONNECTED) && 
		(chan->common.state == WAN_CONNECTED) &&
		(ifp->if_flags & IFF_OACTIVE) != 0){ 
        	/* If our device stays busy for at least 5 seconds then we will
	         * increment statistics. We expect that our device never stays 
        	 * busy more than 5 seconds. 
	         */
		if ((SYSTEM_TICKS - chan->tick_counter) > (5 * hz)){
			ifp->if_collisions ++;
			chan->drvstats_if_send.if_send_tbusy ++;
			chan->drvstats_if_send.if_send_tbusy_timeout ++;
			DEBUG_EVENT("%s: %s: Transmit timed out (%d, %d)\n", 
					card->devname, chan->name, 
					SYSTEM_TICKS, chan->tick_counter);
		} 
	}

	/* For API, the socket family not specified */
	if (chan->usedby != API && dst->sa_family != AF_INET){
		/* protocol family is not supported */
		DEBUG_EVENT("%s:%s: Protocol family (%d) is not isupported ...\n", 
				card->devname, chan->name, dst->sa_family);
		wan_skb_free(m);	/*m_freem(m);*/
		return EAFNOSUPPORT;
	}

	len = m->m_len;
	sendpacket = mtod(m, caddr_t);
	data = mtod(m, caddr_t);

	/* check the udp packet type */
	udp_type = udp_pkt_type(card, m);
	if(udp_type != UDP_INVALID_TYPE){ 
		wan_bpf_report(ifp, m, 1);
        	if(store_udp_mgmt_pkt(udp_type, UDP_PKT_FRM_STACK, card, m, 
						      chan->dlci)){
			DEBUG_UDP("%s:%s: Store UDP packet in private structure.\n", 
					card->devname, chan->name);
			card->hw_iface.set_bit(card->hw, card->intr_perm_off, FR_INTR_TIMER);
			if (udp_type == UDP_FPIPE_TYPE){
				chan->drvstats_if_send.if_send_PIPE_request ++;
			}
		}
		return 0;
	}

	if ((chan->usedby == API) && (len <= sizeof(api_tx_hdr_t))){ 
		/* FIXME: increment some error statistic */
		wan_skb_free(m);	/*m_freem(m);*/
		return 0;
	}
 
	/* FIXME: can we do better than sendpacket[2]? */
	if ((chan->usedby == WANPIPE) && (sendpacket[2] == 0x45)){ 
		/* check to see if the source IP address is a broadcast or */
	        /* multicast IP address */
        	if (chk_bcast_mcast_addr(card, ifp, m)){
			IF_DROP(&ifp->if_snd);
			wan_skb_free(m);	/*m_freem(m);*/
			return 0;
		}
	}

	/* classify the packet before prepanding link-headers */
	WAN_IFQ_CLASSIFY(&ifp->if_snd, m, dst->sa_family, &pktattr);

	if (card->wandev.state != WAN_CONNECTED){ 
		chan->drvstats_if_send.if_send_wan_disconnected ++;
		IF_DROP(&ifp->if_snd);
		wan_skb_free(m);	/*m_freem(m);*/
	}else if (chan->common.state != WAN_CONNECTED){ 
		chan->drvstats_if_send.if_send_dlci_disconnected ++;
		update_chan_state(ifp);
		IF_DROP(&ifp->if_snd);
		wan_skb_free(m);	/*m_freem(m);*/
        }else{ 
		send_data = 1;
		/* If it's an IPX packet */
#ifdef SUPPORT_IPX
		/* FIXME: can we use an IPX header structure for this logic? */
		if ((chan->usedby == WANPIPE) &&
			sendpacket[1] == 0x00 &&
			sendpacket[2] == 0x80 &&
			sendpacket[6] == 0x81 &&
			sendpacket[7] == 0x37) 
		{
			if( chan->enable_IPX ){ 
				switch_net_numbers(sendpacket, chan->network_number, 0);
			}else{ 
				/* FIXME: increment some statistic here! */
				send_data = 0;
			}
		}
#endif
		if (send_data){ 
    			wan_spin_lock_irq(NULL, &s);
			WAN_IFQ_ENQUEUE(&ifp->if_snd, m, &pktattr, error);
			if (error){
				DEBUG_TX("%s: %s: Send queue is full!\n",
					card->devname, chan->name);
				wan_spin_unlock_irq(NULL, &s);
				return error;	
			}
#if 0
 /* !ALTQ */
	
			if (IF_QFULL(&ifp->if_snd)){
				/* send queue is full - drop the packet and exit */
				IF_DROP(&ifp->if_snd);
				wan_spin_unlock_irq(NULL, &s);
				DEBUG_TX("%s: %s: Send queue is full!\n",
					card->devname, chan->name);
				wan_skb_free(m);	/*m_freem(m);*/
				return ENOBUFS;
			}
			IF_ENQUEUE(&ifp->if_snd, m);

#endif	    
			if ((ifp->if_flags & IFF_OACTIVE) == 0){ 
				fr_start(ifp);
			}
			wan_spin_unlock_irq(NULL, &s);
		}
	}
	return (error);
}

/*
 * ============================================================================
 * fr_start() -  called by the system when a packet is added to the 
 *	         output queue for the interface and the device flags
 *	         indicate the device is not currently busy ( IFF_OACTIVE
 *               is clear ).
 *               We will try to transmit the packet. In case there's no
 *  	         available buffer - we will set IFF_OACTIVE, enable 
 *	         transmit interrupt and exit. And then the tx interrupt 
 *	         handler will call if_start() and once its done - it 
 *	         will reset the IFF_OACTIVE.
 */
static void fr_start(struct ifnet * ifp)
{
	fr_channel_t* 		chan = ifp->if_softc;
	sdla_t*	     		card = chan->common.card;
	fr_tx_buf_ctl_t     	frbuf;
	wan_mbox_t* 		mbox = &card->wan_mbox;
	bus_size_t		frbuf_off;
	struct mbuf* 		opkt = NULL;			/* the packet */
	register struct mbuf*	m = NULL;		/* current mbuf */
	unsigned short		data_length;	/* data length */
	unsigned long 		data_addr;
	unsigned char 		attr = 0;
	wan_smp_flag_t		s;	         /* saved int. state */
	int 	   		err = 0;
	char*			data = NULL; 

   	wan_spin_lock_irq(NULL, &s);

	/* We're trying to drain the queue as much as possible here because 
	 * if we manage to get more than one packet into transmit queue     
	 * ( line-down/line-up scenario, or heavy load ) and take only one  
	 * at a time out we get an effective delay ...		      
	 * BF-0001 
	 */ 

	while(!WAN_IFQ_IS_EMPTY(&ifp->if_snd)){

		WAN_IFQ_POLL(&ifp->if_snd, opkt);
		if (opkt == NULL) break;
#if 0
 /* !ALTQ */

	while ((opkt = ifp->if_snd.ifq_head) != NULL){ /* BF-0001 */ 

#endif
		if (!is_tx_ready(card, chan)){
			DEBUG_TX("%s:%s: TR buffer are full. Enable TX INTR for DLCI %d\n", 
					card->devname, chan->name, chan->dlci);
			/* Set interface is active */
			ifp->if_flags |= IFF_OACTIVE;
			/* Enable transmit interrupt */ 
			setup_for_delayed_transmit(ifp);
			/* Set tick_counter */
			chan->tick_counter = SYSTEM_TICKS; 
			chan->drvstats_if_send.if_send_no_bfrs ++;
			wan_spin_unlock_irq(NULL, &s);
			return;
		}
		/* go through mbuf chain to calculate the total length and
		 * copy packet to local buffer.
		 */
		data_length = 0;
		for(m = opkt; m != NULL; m = m->m_next){
			data_length += m->m_len;
		}
		/*
		 * Check that transmit packet length is less than 
		 * MAX_PACKET_SIZE. Copy transmit packet to tx_data.
		 */
		if (data_length > MAX_PACKET_SIZE){
			DEBUG_EVENT("%s:%s: Transmit packet is too long (%d)!\n", 	
				card->devname, chan->name, data_length);
			/* dequeue packet from IP queue */
			WAN_IFQ_DEQUEUE(&ifp->if_snd, opkt);
			if (opkt){
				wan_skb_free(opkt);	/*m_freem(opkt);*/
			}
#if 0
 /* !ALTQ */

			IF_DEQUEUE(&ifp->if_snd, opkt);
			wan_skb_free(opkt);	/*m_freem(opkt);*/

#endif
			continue; 
		}
		data = card->tx_data;
		data_length = 0;
		for(m = opkt; m != NULL; m = m->m_next){
			bcopy(mtod(m, caddr_t), 
				&card->tx_data[data_length],
				m->m_len);
			data_length += m->m_len;
		}

		attr = 0;
		/* 
		 * For an API transmission, get rid of the API header.
		 * For not API, add 2 bytes encapsulation FR header. */ 
		if (chan->usedby == API){ 
			api_tx_hdr_t* api_tx_hdr;
			api_tx_hdr = (api_tx_hdr_t*)card->tx_data;
			attr = api_tx_hdr->attr;
			data = &card->tx_data[sizeof(api_tx_hdr_t)];
			data_length -= sizeof(api_tx_hdr_t);
			err = fr_write(card, chan->dlci, attr, data_length);
		}else{
			err = fr_write(card, chan->dlci, attr, data_length+2);
		}
		if (err){ 
			switch(err){
			case FRRES_DISABLED:
				DEBUG_EVENT("%s:%s: Failed to send data. Communication disabled.\n",
					card->devname, chan->name);
				break;

			case FRRES_INOPERATIVE:
				DEBUG_EVENT("%s:%s: Failed to send data. Channel is inoperative.\n",
					card->devname, chan->name);
				break;

			case FRRES_DLCI_INACTIVE:
				DEBUG_EVENT("%s:%s: Failed to send data. DLCI %d is inactive.\n",
					card->devname, chan->name, chan->dlci);
				break;

			case FRRES_TOO_MANY:
				DEBUG_EVENT("%s:%s: Failed to send data %d bytes frame on DLCI %d\n,",
						card->devname, chan->name, data_length, chan->dlci);
				DEBUG_EVENT("due to it can't be fit in a single Tc period. \n");
				DEBUG_EVENT("Maximum frame size that can fit within Tc period\n");
				DEBUG_EVENT("is %d bytes. Please, either increase CIR\n", 
							mbox->wan_data_len);
				DEBUG_EVENT("for this DLCI or decrease the MTU for device\n"); 
				break;
		
			case FRRES_CIR_OVERFLOW:
			case FRRES_BUFFER_OVERFLOW:
				DEBUG_TX("%s:%s: Failed to send data (no buffers/CIR).\n",
						card->devname, chan->name);
				DEBUG_TX("%s:%s: Enabling tx interrupt for DLCI %d\n", 
						card->devname, chan->name, chan->dlci); 
				/* mark interface is active */
				ifp->if_flags |= IFF_OACTIVE;
	        		setup_for_delayed_transmit(ifp);
				chan->tick_counter = SYSTEM_TICKS; 
				chan->drvstats_if_send.if_send_adptr_bfrs_full++;
				break;

			default:
				DEBUG_EVENT("%s:%s: Failed to send data (0x%02xh)\n", 
						card->devname, chan->name, err);
				chan->drvstats_if_send.if_send_dlci_disconnected++;
				IF_DROP(&ifp->if_snd);
				break;
			}
			wan_spin_unlock_irq(NULL, &s);
			return;
		}
		frbuf_off = *(unsigned long*)mbox->wan_data;
		card->hw_iface.peek(card->hw, frbuf_off, &frbuf, sizeof(frbuf));
		data_addr = frbuf.offset;
		/* 
		 * Add the encapsulation bytes for IP : 03 CC -
		 * tx_elm.flag (only for not API). */
		if (chan->usedby != API){
			unsigned short ip_over_fr_header = FR_IP_HEADER;
			card->hw_iface.poke(card->hw, 
					    data_addr,
					    &ip_over_fr_header,
					    2);
			data_addr += 2;
        	}
		/* Copy local buffer to the board */
		card->hw_iface.poke(card->hw, data_addr, data, data_length);

		/* data length - offset 5, length 2 */
		if (chan->usedby == API){
			frbuf.length = data_length;
		}else{
			frbuf.length = data_length + 2;
		}
		/* dlci number - offset 7, length 2 */
		frbuf.dlci	= chan->dlci;
		/* FECN/BECN bits - offset 9, length 1 */
		frbuf.attr	= 0;
 		card->hw_iface.poke(card->hw,
				    frbuf_off,
				    &frbuf,
				    sizeof(fr_tx_buf_ctl_t));
		/* currently set flag to 0 */
		frbuf.flag = 0x01;
 		card->hw_iface.poke(card->hw,
				    frbuf_off,
				    &frbuf,
				    sizeof(fr_tx_buf_ctl_t));

		wan_bpf_report(ifp, opkt, 1);

		/* dequeue packet from IP queue */
		WAN_IFQ_DEQUEUE(&ifp->if_snd, opkt);
		if (opkt){
			wan_skb_free(opkt);	/*m_freem(opkt);*/
		}
#if 0
 /* !ALTQ */

		IF_DEQUEUE(&ifp->if_snd, opkt);
		wan_skb_free(opkt);	/*m_freem(opkt);*/

#endif
		/* update statistics and free the packet */
		chan->drvstats_if_send.if_send_bfr_passed_to_adptr++;
		ifp->if_opackets++;
        	ifp->if_obytes += data_length;
	} /* while there're packets in the queue */

	/* restore interrupt level and exit */
	wan_spin_unlock_irq(NULL, &s);
	return;
}

/*
 * ============================================================================
 * Setup so that a frame can be transmitted on the occurence of a transmit
 * interrupt.
 */
static void 
setup_for_delayed_transmit (struct ifnet* ifp)
{
	fr_channel_t* 		chan = ifp->if_softc;
	sdla_t* 		card = chan->common.card;
	unsigned int		offset = offsetof(fr_dlci_interface_t, gen_interrupt);

#if 0
TODAY
    if(chan->transmit_length){ 
	log(LOG_ERR, "%s: Big mess in setup_for_del...\n", card->devname);
        return;
    }
    chan->transmit_length = 1;
#endif

	card->hw_iface.set_bit(card->hw, chan->dlci_int_interface_off + offset, FR_INTR_TXRDY);

	/*dlci_interface->packet_length = len; */
	card->hw_iface.set_bit(card->hw, card->intr_perm_off, FR_INTR_TXRDY);

	card->u.f.tx_interrupts_pending ++;
}

/* 
 * ============================================================================
 * Check to see if the packet to be transmitted contains a broadcast or
 * multicast source IP address.
 * Return 0 if not broadcast/multicast address, otherwise return 1.
 */
static int chk_bcast_mcast_addr(sdla_t *card, struct ifnet* ifp,
                                struct mbuf *m)
{
    uint32_t 		src_ip_addr = 0;
    uint32_t 		broadcast_ip_addr = 0;
    struct ifaddr*	ifa = WAN_TAILQ_FIRST(ifp);
    struct sockaddr_in*	saddr = NULL;
    fr_channel_t* 	chan = ifp->if_softc;

    /* read the IP source address from the outgoing packet */
    src_ip_addr = *(uint32_t *)(mtod(m, caddr_t) + 14);

    /* read the IP broadcast address for the device */
    if ((saddr = (struct sockaddr_in *)ifa->ifa_broadaddr) != NULL){ 
        broadcast_ip_addr = htonl(saddr->sin_addr.s_addr);
    }

    /* check if the IP Source Address is a Broadcast address */
    if((ifp->if_flags & IFF_BROADCAST) && (src_ip_addr == broadcast_ip_addr)){ 
        DEBUG_EVENT("%s:%s: Broadcast Source Address silently discarded\n",
					card->devname, chan->name);
        return 1;
     } 

     /* check if the IP Source Address is a Multicast address */
     if ((chan->mc == WANOPT_NO) && (ntohl(src_ip_addr) >= 0xE0000001) && 
		 (ntohl(src_ip_addr) <= 0xFFFFFFFE)){
         DEBUG_EVENT("%s:%s: Multicast Source Address silently discarded\n", 
					card->devname, chan->name);
         return 1;
     }

    return 0;
}

/*
 * ============================================================================
 * Reply to UDP Management system.
 * Return nothing.
 */
static int reply_udp(wan_udp_pkt_t* udp_pkt, unsigned int mbox_len ) 
{
    unsigned char*	data = (unsigned char*)udp_pkt;
    unsigned short	len, udp_length, temp, ip_length;
    unsigned long	ip_temp;
    int			even_bound = 0;

    /* Set length of packet */
    len = /*sizeof(fr_encap_hdr_t)+*/
    	  sizeof(struct ip) + sizeof(struct udphdr)+
          sizeof(wan_mgmt_t) + sizeof(wan_cmd_t) + 
	  mbox_len;

    /* fill in UDP reply */
    udp_pkt->wan_udp_request_reply = UDPMGMT_REPLY;

    /* fill in UDP length */
    udp_length = sizeof(struct udphdr) + 
          	 sizeof(wan_mgmt_t) + sizeof(wan_cmd_t) + 
	  	 mbox_len;

    /* put it on an even boundary */
    if (udp_length & 0x0001){ 
	udp_length += 1;
	len += 1;
	even_bound = 1;
    }

    temp = (udp_length<<8)|(udp_length>>8);
    udp_pkt->wan_udp_len = temp;

    /* swap UDP ports */
    temp = udp_pkt->wan_udp_sport;
    udp_pkt->wan_udp_sport = udp_pkt->wan_udp_dport; 
    udp_pkt->wan_udp_dport = temp;

    /* add UDP pseudo header */
    temp = 0x1100;
    *((unsigned short *)(udp_pkt->wan_udp_data + mbox_len + even_bound)) = temp;	
    temp = (udp_length<<8)|(udp_length>>8);
    *((unsigned short *)(udp_pkt->wan_udp_data + mbox_len + even_bound + 2)) = temp;

    /* calculate UDP checksum */
    udp_pkt->wan_udp_sum = 0;
    udp_pkt->wan_udp_sum = 
		wan_calc_checksum(&data[UDP_OFFSET/*+sizeof(fr_encap_hdr_t)*/],
		udp_length+UDP_OFFSET);

    /* fill in IP length */
    ip_length = udp_length + sizeof(struct ip);
    temp = (ip_length<<8)|(ip_length>>8);
    udp_pkt->wan_ip_len = temp;

    /* swap IP addresses */
    ip_temp = udp_pkt->wan_ip_src;
    udp_pkt->wan_ip_src = udp_pkt->wan_ip_dst;
    udp_pkt->wan_ip_dst = ip_temp;

    /* fill in IP checksum */
    udp_pkt->wan_ip_sum = 0;
    udp_pkt->wan_ip_sum = 
		wan_calc_checksum(&data[0/*sizeof(fr_encap_hdr_t)*/],
	        sizeof(struct ip));

    return len;
} /* reply_udp */

/* 
 * ============================================================================
 * If incoming is 0 (outgoing)- if the net numbers is ours make it 0
 * if incoming is 1 - if the net number is 0 make it ours 
 *
 */
#ifdef SUPPORT_IPX
static void 
switch_net_numbers(unsigned char *sendpacket, unsigned long network_number, 
		   unsigned char incoming)
{
    unsigned long pnetwork_number;

    pnetwork_number = (unsigned long)(
		(sendpacket[14] << 24) + (sendpacket[15] << 16) + 
		(sendpacket[16] << 8) + sendpacket[17]);

    if (!incoming){ 
	/* If the destination network number is ours, make it 0 */
	if( pnetwork_number == network_number){ 
	    sendpacket[14] = sendpacket[15] = sendpacket[16] = 
				 sendpacket[17] = 0x00;
	}
    }else{ 
	/* If the incoming network is 0, make it ours */
	if( pnetwork_number == 0){ 
	    sendpacket[14] = (unsigned char)(network_number >> 24);
	    sendpacket[15] = (unsigned char)((network_number & 
			                      0x00FF0000) >> 16);
	    sendpacket[16] = (unsigned char)((network_number & 
			                      0x0000FF00) >> 8);
	   sendpacket[17] = (unsigned char)(network_number & 
			 		    0x000000FF);
	}
    }

    pnetwork_number = (unsigned long)(
		(sendpacket[26] << 24) + (sendpacket[27] << 16) + 
		(sendpacket[28] << 8) + sendpacket[29]);

    if (!incoming){ 
	/* If the source network is ours, make it 0 */
	if (pnetwork_number == network_number){ 
	    sendpacket[26] = sendpacket[27] = sendpacket[28] = 
			 sendpacket[29] = 0x00;
	}
    }else{ 
        /* If the source network is 0, make it ours */
        if (pnetwork_number == 0){
	    sendpacket[26] = (unsigned char)(network_number >> 24);
	    sendpacket[27] = (unsigned char)((network_number & 
			 		      0x00FF0000) >> 16);
	    sendpacket[28] = (unsigned char)((network_number & 
			 		      0x0000FF00) >> 8);
	    sendpacket[29] = (unsigned char)(network_number & 
			 		     0x000000FF);
	}
    }
} /* switch_net_numbers */
#endif

/****** Interrupt Handlers **************************************************/

/* 
 * ============================================================================
 * S508 frame relay interrupt service routine.
 */
static void fr_isr (sdla_t* card)
{
	fr508_flags_t 	flags;
	wan_mbox_t*	mbox = &card->wan_mbox;
	int 		i,err;

	/* 
	 * This flag prevents nesting of interrupts.  See sdla_isr() routine
	 * in sdlamain.c. 
	 */
	card->in_isr = 1;

	++card->statistics.isr_entry;
	card->hw_iface.peek(card->hw, card->flags_off, &flags, sizeof(flags));
	switch(flags.iflag){
	case FR_INTR_RXRDY:  /* receive interrupt */
		++card->statistics.isr_rx;
	        rx_intr(card);
        	break;

	case FR_INTR_TXRDY:  /* transmit interrupt */
		++ card->statistics.isr_tx; 
		tx_intr(card); 
        	break;

	case FR_INTR_READY:
		Intr_test_counter++;
		++card->statistics.isr_intr_test;
        	break;	

	case FR_INTR_DLC: /* Event interrupt occured */
		WAN_MBOX_INIT(mbox);
		mbox->wan_command = FR_READ_STATUS;
		mbox->wan_data_len = 0;
		mbox->wan_fr_dlci = 0;
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
		if (err){
			fr_event(card, err, mbox);
		}
		break;

	case FR_INTR_TIMER:  /* Timer interrupt */
		timer_intr(card);
		break;

	case FR_INTR_MODEM:
		WAN_MBOX_INIT(mbox);
		mbox->wan_command = FR_READ_STATUS;
		mbox->wan_data_len = 0;
		mbox->wan_fr_dlci = 0;
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
		if (err){
			fr_event(card, err, mbox);
		}
		break;

	default:
		{
			unsigned char str[20];
		 	++card->statistics.isr_spurious;
        		spur_intr(card);
	        	DEBUG_EVENT("%s: Interrupt Type 0x%02x!\n", card->devname, flags.iflag);
			memset(str, ' ', 20);
			for(i = 0; i < 8; i ++){
			    card->hw_iface.peek(card->hw, 
						card->intr_type_off + 0x28 + i, 
						&str[i*2], 
						sizeof(char));
			}
			str[i*2-1] = '\0';
   			DEBUG_EVENT("%s: ID Bytes = %s\n", 
						card->devname, str);	
			break;
		}
	}

	card->in_isr = 0;
	card->hw_iface.poke_byte(card->hw, card->intr_type_off, 0x00);
}


/* 
 * ============================================================================
 * Receive interrupt handler.
 * When a receive interrupt occurs do the following:
 *	1- Find the structure for the dlci that the interrupt occured on
 *      2- If it doesn't exist then print appropriate msg and goto step 8. 
 * 	3- If it exist then copy data to a mbuf.
 * 	4- If mbuf contains Sangoma UDP data then process them
 *	5- If mbuf contains IPXWAN data then send IPXWAN reply packets
 *	6- If mbuf contains Inverse Arp data then send Inv Arp replies
 *	7- If mbuf contains any other data then decapsulate the packet and
 *	   send it to the stack.
 * 	8- Release the receive element and update receive pointers on the board 
 */
static void rx_intr (sdla_t* card)
{
	fr_rx_buf_ctl_t	frbuf;
	fr508_flags_t	flags;
	char		ch;
	fr_channel_t* 	chan = NULL;
	struct mbuf* 	ipkt = NULL;
	struct ifnet* 	ifp = NULL;
	int 		ip_pkt = 0; /* default value */
	unsigned 		dlci, len, offs;
	int 		i, udp_type;	

	card->hw_iface.peek(card->hw, 
				card->rxmb_off, 
				&frbuf, 
				sizeof(fr_rx_buf_ctl_t));
	if (frbuf.flag != 0x01){ 
		card->hw_iface.peek(card->hw, 
					card->flags_off, 
					&flags, 
					sizeof(fr508_flags_t));
		DEBUG_EVENT("%s: corrupted Rx buffer @ 0x%x, flag = 0x%02x!\n", 
					card->devname, 
					(unsigned)card->rxmb_off, 
					frbuf.flag);
		DEBUG_EVENT("%s: ID Bytes = ",card->devname);
		for(i = 0; i < 8; i ++){
			card->hw_iface.peek(card->hw, 
						card->intr_type_off + 0x28 + i,
						&ch,
						1);
			DEBUG_EVENT("0x%02x ", ch);
		}
		DEBUG_EVENT("\n");
		++card->statistics.rx_intr_corrupt_rx_bfr;
		/*
		 * 3.10.2000 RX_CORR
		 * This is case not supposed to be, but if it is, disable
		 * interrupt for current router.
		 */
		DEBUG_EVENT("%s: Critical router failure...\n", card->devname);
		DEBUG_EVENT("Please contact your Sangoma representation!!\n");
		fr_set_intr_mode(card, 0, 0, 0);
		return;
	}
	len  = frbuf.length;
	dlci = frbuf.dlci;
	offs = frbuf.offset;
	/* Find network interface for this packet */
	ifp = find_channel(card, dlci);

	if (ifp == NULL){ 
		/* unconfigured DLCI, so discard packet */
		DEBUG_EVENT("%s: received data on unconfigured DLCI %d!\n",
					card->devname, dlci);
		++card->statistics.rx_intr_on_orphaned_DLCI; 
	}else{ 
		char*		data_ptr = NULL;
		unsigned int 	api_hdr_len = 0;
		unsigned int 	hdr_len = 0;

		chan = ifp->if_softc;

		if ((ifp->if_flags & (IFF_UP|IFF_RUNNING)) != (IFF_UP|IFF_RUNNING)){ 
			DEBUG_EVENT("%s: RX packet on DLCI %d however device %s is not UP and RUNNING.\n",
					card->devname, dlci, chan->name);
			chan->drvstats_rx_intr.rx_intr_dev_not_started ++;
			ifp->if_iqdrops ++;
			goto rx_intr_done;
		}

		if (chan->usedby == API){
			api_hdr_len = sizeof(api_rx_hdr_t);
		}

		/*
		 * Check that packet length is less then MAX_PACKET_SIZE.
		 */
		if (len + api_hdr_len > MAX_PACKET_SIZE){
			DEBUG_EVENT("%s:%s: RX packet is too long!\n", 
					card->devname, chan->name);
			goto rx_intr_done; 
		}

		/*
		 * Now, we can start copying data into data area. 
		 * We len bytes. Note that since the S508s receive buffers are  
		 * cyclic we might have to copy the packet in two pieces. 
		 */
		if ((offs + len) <= card->u.f.rx_top_off + 1){ 
			card->hw_iface.peek(card->hw, 
					offs, 
					&card->rx_data[api_hdr_len],
					len);
			offs += len;
		}else{
			unsigned int first = card->u.f.rx_top_off - offs + 1;
			unsigned int second = len - first;
			card->hw_iface.peek(card->hw, 
					offs, 
					&card->rx_data[api_hdr_len], 
					first);
			card->hw_iface.peek(card->hw, 
					card->u.f.rx_base_off, 
					&card->rx_data[api_hdr_len + first],
					second); 
			offs = card->u.f.rx_base_off + second;
		}

		if (chan->usedby == API){ /* Update API header */
			((api_rx_hdr_t*)card->rx_data)->attr = frbuf.attr;
			((api_rx_hdr_t*)card->rx_data)->time_stamp = frbuf.tmstamp;
		}

		/* check the the packet type ... */
		if (chan->usedby != API){
			if (card->rx_data[api_hdr_len] == chan->fr_encap_0){ 
		    		/* IP over Frmae Relay encapsulation ... 
				 * skip first two bytes and
				 * go on with mbufs, UDP stuff and so on ...
				 */
				len -= 2; 
				hdr_len = 2; 
				ip_pkt = 1;
			}else{
				switch(card->rx_data[api_hdr_len]){
				case CISCO_UI:
					DEBUG_EVENT("%s%s: Changing data encapsulation to CISCO\n",
							card->devname, 
							chan->name);
					chan->fr_encap_0 = CISCO_UI;
					chan->fr_encap_1 = CISCO_IP;
					len -= 2;
					hdr_len = 2;
					ip_pkt = 1;
					break;

				case Q922_UI:
					DEBUG_EVENT("%s:%s: Changing data encapsulation to IETF\n",
							card->devname,
							chan->name);
					chan->fr_encap_0 = Q922_UI;
					chan->fr_encap_1 = NLPID_IP;
					len -= 2;
					hdr_len = 2;
					ip_pkt = 1;
					break;
				
				default:
			
					if (card->rx_data[api_hdr_len] != CISCO_INV){ 
						DEBUG_EVENT("%s: Rx non IETF/CISCO pkt : 0x%02X 0x%02X, dropping !\n",
							card->devname,
							card->rx_data[api_hdr_len],
							card->rx_data[api_hdr_len+1]);
					}
					++chan->drvstats_rx_intr.rx_intr_bfr_not_passed_to_stack;
					++card->wandev.stats.rx_errors;
					ifp->if_iqdrops ++;
					goto rx_intr_done;
					break;
				}
			}
		}
	
		data_ptr = &card->rx_data[hdr_len];
		len += api_hdr_len;

		ipkt = sdla_get(ifp, data_ptr, len);
		if (ipkt == NULL){
			DEBUG_EVENT("%s:%s: Failed to allocate mbuf (RX_INTR)!!\n", 
				card->devname, chan->name);
			ifp->if_iqdrops ++;
			goto rx_intr_done;
		}

		udp_type = udp_pkt_type(card, ipkt);
	        if (udp_type != UDP_INVALID_TYPE){ 
			wan_bpf_report(ifp, ipkt, 1);
			if (store_udp_mgmt_pkt(udp_type,UDP_PKT_FRM_NETWORK, card, 
			  ipkt, dlci)){ 
				DEBUG_UDP("%s:%s: RX UDP packet from remote end.\n", 
					card->devname, chan->name);
				card->hw_iface.set_bit(card->hw, card->intr_perm_off, FR_INTR_TIMER);
				if (udp_type == UDP_FPIPE_TYPE){
					chan->drvstats_rx_intr.rx_intr_PIPE_request ++;
				}
			}			
		}else{
			int pass_to_stack = 1;

			if (chan->usedby != API){
#ifdef SUPPORT_IPX 
				if (handle_IPXWAN(mtod(ipkt, caddr_t), chan->name,	
					          chan->enable_IPX, chan->network_number)){ 
					if (chan->enable_IPX){
						fr_send(card, dlci, 0, ipkt->m_len, 
								       mtod(ipkt, caddr_t), 1);
					}
					wan_skb_free(ipkt);	/*m_freem(ipkt);*/
					pass_to_stack = 0;
				}else 
#endif
					if (is_arp(mtod(ipkt, caddr_t))) { 
						if (process_ARP((arphdr_1490_t *)mtod(ipkt, caddr_t), 
									card, ifp)){
							DEBUG_EVENT("%s: Error processing ARP Packet.\n", 
									card->devname);
						}
					    	wan_skb_free(ipkt);	/*m_freem(ipkt);*/
					        pass_to_stack = 0;
					}else{ 
						if (!ip_pkt){ 
							DEBUG_EVENT("%s:%s: Non IETF/CISCO packet discarded\n", 
									card->devname, chan->name);		
							wan_skb_free(ipkt);	/*m_freem(ipkt);*/
							pass_to_stack = 0;
						}
					}
			}
			if (pass_to_stack){
#if defined(__NetBSD__) || defined(__OpenBSD__) || (__FreeBSD_version < 501000)
				wan_smp_flag_t	s;
#endif
				wan_bpf_report(ifp, ipkt, 1);
				/* 
				 * Schedule IP input routine. Note that it makes 
				 * much more sence to check first if the protocol 
				 * queue is full - and then to schedule the input 
				 * routine, but i took a look at kernel sources
				 * for ether_input() 
				 * [ /usr/src/sys/net/if_ethersubr.c ] and they
				 * do it in this order - so i figured they know 
				 * better.
				 */
#if (__FreeBSD_version >= 501000)
				netisr_queue(NETISR_IP, ipkt);
#else
   				wan_spin_lock_irq(NULL, &s);
				schednetisr(NETISR_IP);
				if (IF_QFULL(&ipintrq)){ 
					/* oh, no - IP queue is full - well - we'll 
					 * try again later 
					 */
					DEBUG_RX("%s:%s: IP queue is full, drop packet!\n",
							card->devname, chan->name);
					IF_DROP(&ipintrq);
					wan_skb_free(ipkt);	/*m_freem(ipkt);*/
					wan_spin_unlock_irq(NULL, &s);
					goto rx_intr_done;
				}
  
				/* ok enqueue the packet */
				IF_ENQUEUE(&ipintrq, ipkt);
				wan_spin_unlock_irq(NULL, &s);
#endif
				/* update interface statistics */
				ifp->if_ipackets++;
				ifp->if_ibytes += len;
				chan->drvstats_rx_intr.rx_intr_bfr_passed_to_stack ++;
			} 
		} 
	}

rx_intr_done:

	/* Release buffer element and calculate a pointer to the next one */ 
	frbuf.flag = 0x00;
	card->hw_iface.poke(card->hw, card->rxmb_off, &frbuf, sizeof(frbuf));
	card->rxmb_off += sizeof(fr_rx_buf_ctl_t);
	if (card->rxmb_off > card->u.f.rxmb_last_off){
		card->rxmb_off = card->u.f.rxmb_base_off;
	}
}

/* 
 * ============================================================================
 * Transmit interrupt handler.
 */
static void tx_intr(sdla_t *card)
{
    fr508_flags_t	flags;
    fr_tx_buf_ctl_t	bctl;
    unsigned long	tse_offset;
    struct ifnet* ifp = NULL;
    fr_channel_t* chan = NULL;

    card->hw_iface.peek(card->hw, card->flags_off, &flags, sizeof(fr508_flags_t));
    tse_offset = flags.tse_offs;
    /* Find the structure and make it unbusy */
    ifp = find_channel(card, flags.dlci);
    if (ifp == NULL){
	/* no suppose to happened - but just to check */
	DEBUG_EVENT("%s: Tx interrupt for DLCI %d which is not configured.\n", 
					card->devname, flags.dlci);
	return;
    }
    chan = ifp->if_softc;

#if 0
TODAY
    if(!chan->transmit_length) 
    {
	log(LOG_ERR, "%s: tx int error - transmit length zero\n", 
						card->wandev.name);
        return;
    }
#endif   
    {
	wan_smp_flag_t	s;
	int 		transmit = 1;
	struct mbuf*	opkt = NULL, *m = NULL;
	unsigned long 	data_addr;
	unsigned short 	data_length = 0;

        wan_spin_lock_irq(NULL, &s);

	/* check if we've got anything to tranmit at all */
	if (WAN_IFQ_IS_EMPTY(&ifp->if_snd)){
#if 0
 /* !ALTQ */
	if ((opkt = ifp->if_snd.ifq_head) == NULL){ 
#endif
	    /* not supposed to happen - but who knows */
	    DEBUG_EVENT("%s:%s: Nothing to transmit on DLCI %d.\n", 
				card->devname, chan->name, flags.dlci);
	    /* set up the no-transmit flag */
	    transmit = 0;
	}else{ 
	    /* ok, now we actually gonna copy the data onto the board and   */
	    /* prepare the rest of tx status element (length only actually) */

	    /* dequeue packet from IP queue */
	    WAN_IFQ_DEQUEUE(&ifp->if_snd, opkt);
	    if (opkt == NULL){
		transmit = 0;
		goto transmit_start;
	    }
#if 0
 /* !ALTQ */
	    IF_DEQUEUE(&ifp->if_snd, opkt);
#endif

	    /* initialize some of our stuff */
	    card->hw_iface.peek(card->hw, tse_offset, &bctl, sizeof(fr_tx_buf_ctl_t));
	    data_addr = bctl.offset; 
	    data_length	= 0;

	    /* Add the encapsulation bytes for IP  */
	    {
	        unsigned short ip_over_fr_header = FR_IP_HEADER;

	        card->hw_iface.poke(card->hw, data_addr, &ip_over_fr_header, 2 );
	        data_addr += 2; 
	        data_length += 2;
	    }

	    /* go through mbuf chain and copy data to the board */
	    for(m = opkt; m != NULL; m = m->m_next){ 
		/* copy data to the board */
	        card->hw_iface.poke(card->hw, data_addr, mtod(m, caddr_t), m->m_len );
		data_length += m->m_len;
		data_addr += m->m_len;
	    }
	}

transmit_start:
	/* see if we need to transmit packet - if yes - set the flag to 0x80 
         * if not - set the flag to 0xa0				  
	 */
	if (transmit){ 
	    /* 
	     * set up the rest of tx element and transmit the frame ....
	     */
	    card->hw_iface.peek(card->hw, tse_offset, &bctl, sizeof(fr_tx_buf_ctl_t));
	    bctl.dlci   = flags.dlci;
            bctl.length = data_length;
	    bctl.attr   = 0;
	    bctl.flag   = 0x0;	/* flag will be set after coping structure to memory */
	    card->hw_iface.poke(card->hw, tse_offset, &bctl, sizeof(fr_tx_buf_ctl_t));
	    bctl.flag = 0xC0;
	    card->hw_iface.poke(card->hw, tse_offset, &bctl, sizeof(fr_tx_buf_ctl_t));
	    wan_bpf_report(ifp, opkt, 1);

	    /* update statistics and free the packet */
	    ifp->if_opackets++;
	    ifp->if_obytes += data_length;
	    wan_skb_free(opkt);	/*m_freem(opkt);*/

	    /* reset the ACTIVE flag */
	    ifp->if_flags &= ~IFF_OACTIVE;

	    /* Update tick_counter */
	    chan->tick_counter = SYSTEM_TICKS;

       	    chan->transmit_length = 0;

            /* if any other interfaces have transmit interrupts pending, 
	     * do not disable the global transmit interrupt 
	     */
	    if (!(-- card->u.f.tx_interrupts_pending)){
		card->hw_iface.clear_bit(card->hw, card->intr_perm_off, FR_INTR_TXRDY);
	    }
	}else{
	    bctl.flag = 0xA0;
	    card->hw_iface.poke(card->hw, tse_offset, &bctl, sizeof(fr_tx_buf_ctl_t));
	}
	wan_spin_unlock_irq(NULL, &s);
    }
}


/*
 * ============================================================================
 * Timer interrupt handler.
 * FIXME: update comments as we modify the code
 * The timer interrupt is used for three purposes:
 *    1) Processing udp calls from 'fpipemon'.
 *    2) Processing update calls from /proc file system
 *    2) Reading board-level statistics for updating the proc file system.
 *    3) Sending inverse ARP request packets.
 */
static void timer_intr(sdla_t *card)
{
	if(card->u.f.timer_int_enabled & TMR_INT_ENABLED_UDP){ 
		if ((card->u.f.udp_type == UDP_FPIPE_TYPE) && 
        	    (process_udp_mgmt_pkt(card))){
        		card->u.f.timer_int_enabled &= ~TMR_INT_ENABLED_UDP;
		}
	}

#ifndef FR_OLD_INTR_TEST
	if (card->u.f.timer_int_enabled & TMR_INT_ENABLED_TEST){ 
		Intr_test_counter++;
    		if (Intr_test_counter >= MAX_NEW_INTR_TEST_COUNTER){ 
    			DEBUG_EVENT("%s: Interrupt Test Passed, Counter: %d\n",
					card->devname, Intr_test_counter);
			card->u.f.timer_int_enabled &= ~TMR_INT_ENABLED_TEST;

			card->hw_iface.clear_bit(card->hw, 
						    card->intr_perm_off, 
						    FR_INTR_TIMER);
			card->hw_iface.set_bit(card->hw, 
						   card->intr_perm_off, 
						   (FR_INTR_RXRDY | FR_INTR_DLC | 
						   FR_INTR_TX_MULT_DLCIs | FR_INTR_MODEM));
 
			fr_comm_enable(card);
			wanpipe_set_state(card, WAN_CONNECTED);
		}
	}
#endif

	if (card->u.f.timer_int_enabled & TMR_INT_ENABLED_UPDATE){ 
		fr_link_stat_t link_stat;
		fr_comm_stat_t err_stats;
		if (!fr_get_err_stats(card, &err_stats)){
			card->wandev.stats.rx_over_errors    = err_stats.rx_overruns;
			card->wandev.stats.rx_crc_errors     = err_stats.rx_bad_crc;
			card->wandev.stats.rx_missed_errors  = err_stats.rx_aborts;
			card->wandev.stats.rx_length_errors  = err_stats.rx_too_long;
			card->wandev.stats.tx_aborted_errors = err_stats.tx_aborts;
		}
		fr_get_stats(card, 0, (void*)&link_stat);
		if (IS_TE1_CARD(card)){
			/* TE1 Update T1/E1 alarms */
			card->wandev.fe_iface.read_alarm(&card->fe, 0);
			/* TE1 Update T1/E1 perfomance counters */
			card->wandev.fe_iface.read_pmon(&card->fe); 
		}else if (IS_56K_CARD(card)){
			/* 56K Update CSU/DSU alarms */
			card->wandev.fe_iface.read_alarm(&card->fe, 1);
		}	
		card->update_comms_stats = 0;
		card->u.f.timer_int_enabled &= ~TMR_INT_ENABLED_UPDATE;
	}

	/* triggered from update() */
	if (card->u.f.timer_int_enabled & TMR_INT_ENABLED_UPDATE_DLCI){
		fr_channel_t*	chan = card->u.f.update_dlci;
		fr_dlci_stat_t	dlci_stat;
	
		fr_get_stats(card, chan->dlci, (void*)&dlci_stat);
		fr_get_dlci_stats(card, chan);
		card->update_comms_stats = 0;
		card->u.f.update_dlci = NULL;
		card->u.f.timer_int_enabled &= ~TMR_INT_ENABLED_UPDATE_DLCI;
	}

	if (card->u.f.timer_int_enabled & TMR_INT_ENABLED_TE){ 
		card->wandev.fe_iface.polling(&card->fe);
		card->u.f.timer_int_enabled &= ~TMR_INT_ENABLED_TE;
	}

	if (card->u.f.timer_int_enabled & TMR_INT_ENABLED_INARP){ 
		struct wan_dev_le	*devle;
		fr_channel_t		*chan = NULL;
		netdevice_t		*ifp = NULL;
	
		WAN_LIST_FOREACH(devle, &card->wandev.dev_head, dev_link){
			ifp = WAN_DEVLE2DEV(devle);
			if (!ifp || !wan_netif_priv(ifp)) continue;
			chan = wan_netif_priv(ifp);
			if (chan->inarp == INARP_REQUEST && 
			    chan->common.state == WAN_CONNECTED &&
			    chan->dlci_configured == DLCI_CONFIGURED){
				send_inarp_request(card, ifp);
			}
		}
		card->u.f.timer_int_enabled &= ~TMR_INT_ENABLED_INARP;
	}

	if (!card->u.f.timer_int_enabled){
		card->hw_iface.clear_bit(card->hw, card->intr_perm_off, FR_INTR_TIMER);
	}
}


/*============================================================================
 * Spurious interrupt handler.
 * o print a warning
 * o 
 */
static void spur_intr (sdla_t* card)
{
    DEBUG_EVENT("%s: spurious interrupt!\n", card->devname);
}


/*===========================================================================
 *  Return 0 for non-IPXWAN packet
 *         1 for IPXWAN packet or IPX is not enabled!
 *  FIXME: Use a IPX structure here not offsets
 */
#ifdef SUPPORT_IPX
static int 
handle_IPXWAN(unsigned char *sendpacket, char *devname, 
	      unsigned char enable_IPX, unsigned long network_number)
{
    int i;

    if (sendpacket[1] == 0x00 &&
        sendpacket[2] == 0x80 &&
        sendpacket[6] == 0x81 &&
        sendpacket[7] == 0x37) 
    { 
	/* It's an IPX packet */
	if (!enable_IPX){
	    /* Return 1 so we don't pass it up the stack. */
	    return 1;
	}
    }else{ /* It's not IPX so return and pass it up the stack. */
	return 0;
    }

    if (sendpacket[24] == 0x90 && sendpacket[25] == 0x04){
	/* It's IPXWAN */
	if (sendpacket[10] == 0x02 && sendpacket[42] == 0x00){
	    /* It's a timer request packet */
	    DEBUG_EVENT("%s: Received IPXWAN Timer Request packet\n",
					devname);

	    /* Go through the routing options and answer no to every
	     * option except Unnumbered RIP/SAP
	     */
	    for(i = 49; sendpacket[i] == 0x00; i += 5){
		/* 0x02 is the option for Unnumbered RIP/SAP */
		if (sendpacket[i + 4] != 0x02)
		    sendpacket[i + 1] = 0;
	    }

	    /* Skip over the extended Node ID option */
	    if (sendpacket[i] == 0x04){
		i += 8;
	    }

	    /* We also want to turn off all header compression opt.
	     */
	    for(; sendpacket[i] == 0x80 ;){
	        sendpacket[i + 1] = 0;
	        i += (sendpacket[i + 2] << 8) + (sendpacket[i + 3]) + 4;
	    }

	    /* Set the packet type to timer response */
	    sendpacket[42] = 0x01;

	    DEBUG_EVENT("%s: Sending IPXWAN Timer Response\n", devname);
        }else{ 
            if (sendpacket[42] == 0x02){
	        /* This is an information request packet */
                DEBUG_EVENT("%s: Received IPXWAN Information Request packet\n",
							devname);

		/* Set the packet type to information response */
		sendpacket[42] = 0x03;

		/* Set the router name */
		sendpacket[59] = 'F';
		sendpacket[60] = 'P';
		sendpacket[61] = 'I';
		sendpacket[62] = 'P';
		sendpacket[63] = 'E';
		sendpacket[64] = '-';
		sendpacket[65] = CVHexToAscii(network_number >> 28);
	        sendpacket[66] = 
			CVHexToAscii((network_number & 0x0F000000) >> 24);
	        sendpacket[67] = 
			CVHexToAscii((network_number & 0x00F00000) >> 20);
	        sendpacket[68] = 
			CVHexToAscii((network_number & 0x000F0000) >> 16);
		sendpacket[69] = 
			CVHexToAscii((network_number & 0x0000F000) >> 12);
	        sendpacket[70] = 
			CVHexToAscii((network_number & 0x00000F00) >> 8);
		sendpacket[71] = 
			CVHexToAscii((network_number & 0x000000F0) >> 4);
	 	sendpacket[72] = 
			CVHexToAscii(network_number & 0x0000000F);
		for(i = 73; i < 107; i+= 1)
		    sendpacket[i] = 0;

		DEBUG_EVENT("%s: Sending IPXWAN Information Response packet\n",
							devname);
	    }else{
		DEBUG_EVENT("%s: Unknown IPXWAN packet!\n",devname);
		return 0;
	    }
	}

        /* Set the WNodeID to our network address */
        sendpacket[43] = (unsigned char)(network_number >> 24);
	sendpacket[44] = (unsigned char)((network_number & 0x00FF0000) >> 16);
        sendpacket[45] = (unsigned char)((network_number & 0x0000FF00) >> 8);
	sendpacket[46] = (unsigned char)(network_number & 0x000000FF);
	return 1;
    }

    /* If we get here, its an IPX-data packet so it'll get passed up the 
     * stack.
     * switch the network numbers 
     */
    switch_net_numbers(sendpacket, network_number ,1);
    return 0;
}
#endif

/*
 * ============================================================================
 * Process Route.
 * This routine is called as a polling routine to dynamically add/delete routes
 * negotiated by inverse ARP.  It is in this "task" because we don't want routes
 * to be added while in interrupt context.
 */
static void process_route (netdevice_t* ifp)
{
    fr_channel_t*	chan = NULL;
    sdla_t*		card = NULL;
    int 		err = 0;
    struct ifaddr*	ifa = WAN_TAILQ_FIRST(ifp);
    struct sockaddr_in*	saddr = NULL;		/* addr. struct */

    WAN_ASSERT1(ifp == NULL);
    WAN_ASSERT1(ifp->if_softc == NULL);
    chan = ifp->if_softc;
    card = chan->common.card;
    /* i have no idea why but it seems that the local and remote IP's 
     * are stored not inside ifa but inside ifa->ifa_next. We'd like  
     * to make sure we are not writing to NULL though.		
     */
    if ((ifa = WAN_TAILQ_NEXT(ifa)) == NULL) 
    {
        /* that shouldnt happen ... */
        return;
    }

    /* delete old routing entry */
    err = rtrequest((int)RTM_DELETE, 		/* action */
		ifa->ifa_dstaddr,		/* destination */
		ifa->ifa_addr,		    /* gateway */
		ifa->ifa_netmask,		/* mask */
		RTF_HOST,			    /* flag */
		(struct rtentry **)0);

    switch(chan->route_flag){ 
    case ADD_ROUTE:
        /* set up new destination address */
        saddr = (struct sockaddr_in *)ifa->ifa_dstaddr;
        saddr->sin_addr.s_addr = chan->dst_addr;

        /* add new routing entry */
        err = rtrequest((int)RTM_ADD, 	  /* action */
		ifa->ifa_dstaddr,	  /* destination */
		ifa->ifa_addr,		  /* gateway */
		ifa->ifa_netmask,	  /* mask */
		RTF_HOST,			  /* flag */
		(struct rtentry **)0);
	if (err){
	    DEBUG_EVENT("%s: %s: Adding of route failed (%d)\n",
				card->devname, chan->name, err);
	    DEBUG_EVENT("%s: %s: Address %s\n", 
				card->devname, chan->name,
				in_ntoa(saddr->sin_addr.s_addr));
	}else{
            chan->route_flag = ROUTE_ADDED;
	}
        break;

    case REMOVE_ROUTE:
	if (err){
	    DEBUG_EVENT("%s: %s: Deleting of route failed (%d)\n",
				card->devname, chan->name, err);
	    DEBUG_EVENT("%s: %s: Address %s\n", 
				card->devname, chan->name,
				in_ntoa(((struct sockaddr_in*)ifa->ifa_dstaddr)->sin_addr.s_addr));
	}else{
            chan->route_flag = NO_ROUTE;
	}
    	break;
    }

}


/****** Frame Relay Firmware-Specific Functions *****************************/

/* 
 * ============================================================================
 * Read firmware code version.
 * o fill string str with firmware version info. 
 */
static int fr_read_version (sdla_t* card, char* str)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int 	retry = MAX_CMD_RETRY;
	int 	err;

	DEBUG_CMD("%s: Read firmware version info ...\n", card->devname);
	WAN_MBOX_INIT(mbox);
	do{
		mbox->wan_command = FR_READ_CODE_VERSION;
		mbox->wan_data_len = 0;
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	} while (err && retry-- && !fr_event(card, err, mbox));

	if (!err && str){ 
		memcpy(str, mbox->wan_data, mbox->wan_data_len);
        	str[mbox->wan_data_len] = '\0';
	}
	return err;
}


/* 
 * ============================================================================
 * Set global configuration.
 */
static int fr_configure (sdla_t* card, fr_conf_t *conf)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int 	retry = MAX_CMD_RETRY;
	int 	dlci_num = card->u.f.dlci_num;
	int 	err;

	DEBUG_CMD("%s: Set global configuration ...\n", card->devname);
	WAN_MBOX_INIT(mbox);
	do{ 
		mbox->wan_command = FR_SET_CONFIG;
		mbox->wan_data_len = sizeof(fr_conf_t) + dlci_num * sizeof(unsigned short);
		memcpy(mbox->wan_data, conf, sizeof(fr_conf_t));
		if (dlci_num){
			int i = 0;
			for (i = 0; i < dlci_num; ++i){
				((fr_conf_t*)mbox->wan_data)->dlci[i] = 
						card->u.f.node_dlci[i];
			}
		}
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	} while (err && retry-- && !fr_event(card, err, mbox));
	return err;
}

/* 
 * ============================================================================
 * Set DLCI configuration.
 */
static int fr_dlci_configure (sdla_t* card, fr_dlc_conf_t *conf, unsigned dlci)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int 	retry = MAX_CMD_RETRY;
	int 	err;

	DEBUG_CMD("%s: Configure DLCI %d...\n", card->devname, dlci);
	WAN_MBOX_INIT(mbox);
	do{ 
		mbox->wan_fr_dlci = (unsigned short) dlci; 
		mbox->wan_command = FR_SET_CONFIG;
		mbox->wan_data_len = sizeof(fr_dlc_conf_t);
		memcpy(mbox->wan_data, conf, sizeof(fr_dlc_conf_t));
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	} while (err && retry--);

	return err;
}


/* 
 * ============================================================================
 * Set interrupt mode.
 */
static int 
fr_set_intr_mode (sdla_t* card, unsigned mode, unsigned mtu, 
				  unsigned short timeout)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	fr508_intr_ctl_t*	ictl = (fr508_intr_ctl_t*)mbox->wan_data;
	int 		retry = MAX_CMD_RETRY;
	int 		err;

	DEBUG_CMD("%s: Set interrupt mode...\n", card->devname);
	WAN_MBOX_INIT(mbox);
	do{
		mbox->wan_data_len = sizeof(fr508_intr_ctl_t);
		mbox->wan_command = FR_SET_INTR_MODE;

		bzero(ictl, sizeof(fr508_intr_ctl_t));
		ictl->mode   = mode;
		ictl->tx_len = mtu;
		ictl->irq    = card->wandev.irq;
		/* indicate timeout on timer */
		if (mode & FR_INTR_TIMER) ictl->timeout = timeout; 
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	} while (err && retry-- && !fr_event(card, err, mbox));

	return err;
}

/* 
 * ============================================================================
 * Enable communications.
 */
static int fr_comm_enable (sdla_t* card)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int 	retry = MAX_CMD_RETRY;
	int 	err;

	DEBUG_CMD("%s: Enable communications...\n", card->devname);
	WAN_MBOX_INIT(mbox);
	do{
		mbox->wan_command = FR_COMM_ENABLE;
		mbox->wan_data_len = 0;
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	} while (err && retry-- && !fr_event(card, err, mbox));

	return err;
}

/*
 * ============================================================================
 * Disable communications. 
 */
static int fr_comm_disable (sdla_t* card)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int 		retry = MAX_CMD_RETRY;
	int 		err;

	DEBUG_CMD("%s: Disable communications...\n", card->devname);
	WAN_MBOX_INIT(mbox);
	do{
		mbox->wan_command = FR_COMM_DISABLE;
		mbox->wan_data_len = 0;
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	} while (err && retry-- && !fr_event(card, err, mbox));

	retry = MAX_CMD_RETRY;
	WAN_MBOX_INIT(mbox);
	do{
		mbox->wan_command = FR_SET_MODEM_STATUS;
		mbox->wan_data_len = 1;
		mbox->wan_data[0] = 0x00;
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	} while (err && retry-- && !fr_event(card, err, mbox));

	return err;
}

/*
 * ============================================================================
 * Get communications error statistics. 
 */
static int fr_get_err_stats (sdla_t* card, fr_comm_stat_t* stats)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int		retry = MAX_CMD_RETRY;
	int		err;

	DEBUG_CMD("%s: Get error statistics...\n", card->devname);
	WAN_MBOX_INIT(mbox);
	do{ 
		mbox->wan_command = FR_READ_ERROR_STATS;
		mbox->wan_data_len = 0;
		mbox->wan_fr_dlci = 0;
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	} while (err && retry-- && !fr_event(card, err, mbox));

	if (!err){ 
		memcpy(stats, mbox->wan_data, sizeof(fr_comm_stat_t));
	}

	return err;
}


/* 
 * ============================================================================
 * Get statistics. 
 */
static int fr_get_stats (sdla_t* card, int dlci, void* stats)
{
    	wan_mbox_t*	mbox = &card->wan_mbox;
	int		retry = MAX_CMD_RETRY;
    	int		err;

    	DEBUG_CMD("%s: Get statistics...\n", card->devname);
	WAN_MBOX_INIT(mbox);
    	do{
		mbox->wan_command = FR_READ_STATISTICS;
		mbox->wan_data_len = 0;
		mbox->wan_fr_dlci = dlci;
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
    	} while (err && retry-- && !fr_event(card, err, mbox));

    	if (!err){ 
		if (dlci == 0){
			/* Global statistics */
	    		fr_link_stat_t* link_stat = (fr_link_stat_t*)stats;
			memcpy(link_stat, mbox->wan_data,  sizeof(fr_link_stat_t));
	    		card->wandev.stats.rx_frame_errors = 
				link_stat->rx_bad_format;
	    		card->wandev.stats.rx_dropped =
				link_stat->rx_dropped + link_stat->rx_dropped2;
		}else{
	    		/* Specific statistic for DLCI */
			memcpy(stats, mbox->wan_data,  sizeof(fr_dlci_stat_t));
		}
    	}

    	return err;
}

/* 
 * ============================================================================
 * Get statistics. 
 */
static int fr_get_dlci_stats (sdla_t* card, fr_channel_t* chan)
{
    	wan_mbox_t*	mbox = &card->wan_mbox;
	int		retry = MAX_CMD_RETRY;
	int		dlci = 0;
    	int		err = 0;

	if (chan){
		dlci = chan->dlci;
	}

	WAN_MBOX_INIT(mbox);
    	if (dlci){
       		do{
	    		mbox->wan_command = FR_READ_ADD_DLC_STATS;
	    		mbox->wan_data_len = 0;
	    		mbox->wan_fr_dlci = dlci;
			err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
        	} while (err && retry-- && !fr_event(card, err, mbox));
		if (!err){
	    		struct fr_stats { 
				unsigned short FECN; 
				unsigned short BECN; 
			} stats;
			memcpy(&stats, mbox->wan_data, sizeof(struct fr_stats));
	    		chan->rx_FECN = stats.FECN;
	    		chan->rx_BECN = stats.BECN;
		}
    	}
  
    	return err;
}

/*
 * ============================================================================
 * Add DLCI(s) (Access Node only!).
 * This routine will perform the ADD_DLCIs command for the specified DLCI.
 */
static int fr_add_dlci (sdla_t* card, int dlci)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int		retry = MAX_CMD_RETRY;
	int		err;

	DEBUG_CMD("%s: Add new DLCI %d...\n", card->devname, dlci);
	WAN_MBOX_INIT(mbox);
	do{
		mbox->wan_data_len  = sizeof(short);
		mbox->wan_command = FR_ADD_DLCI;
		memcpy(mbox->wan_data, (void*)&dlci, sizeof(short));
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	} while (err && retry-- && !fr_event(card, err, mbox));

	return err;
}

/* 
 * ===========================================================================
 * Activate DLCI(s) (Access Node only!). 
 * This routine will perform the ACTIVATE_DLCIs command with a DLCI number. 
 */
static int fr_activate_dlci (sdla_t* card, int dlci)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int retry = MAX_CMD_RETRY;
	int err;

	DEBUG_CMD("%s: Active DLCI %d...\n", card->devname, dlci);
	WAN_MBOX_INIT(mbox);
	do{ 
		mbox->wan_data_len  = sizeof(short);
		mbox->wan_command = FR_ACTIVATE_DLCI;
		memcpy(mbox->wan_data, (void*)&dlci, sizeof(short));
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	} while (err && retry-- && !fr_event(card, err, mbox));

	return err;
}

/* 
 * ============================================================================
 * Delete DLCI(s) (Access Node only!). 
 * This routine will perform the DELETE_DLCIs command with a DLCI number. 
 */
static int fr_delete_dlci (sdla_t* card, int dlci)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int retry = MAX_CMD_RETRY;
	int err;

	DEBUG_CMD("%s: Delete DLCI %d...\n", card->devname, dlci);
	WAN_MBOX_INIT(mbox);
	do{
		mbox->wan_data_len  = sizeof(short);
		mbox->wan_command = FR_DELETE_DLCI;
		memcpy(mbox->wan_data, (void*)&dlci, sizeof(short));
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	} while (err && retry-- && !fr_event(card, err, mbox));

	return err;
}


/*
 * ============================================================================
 * Issue in-channel signalling frame. 
 */
static int fr_issue_isf (sdla_t* card, int isf)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int retry = MAX_CMD_RETRY;
	int err;

	DEBUG_CMD("%s: Issue signalling frame...\n", card->devname);
	WAN_MBOX_INIT(mbox);
	do{
		mbox->wan_data_len  = 1;
		mbox->wan_command = FR_ISSUE_IS_FRAME;
		mbox->wan_data[0] = isf;
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	} while (err && retry-- && !fr_event(card, err, mbox));

	return err;
}


/*
 * ===========================================================================
 * Send data to the card... 
 */
static int fr_write(sdla_t *card, int dlci, unsigned char attr, int len)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	int		retry = MAX_CMD_RETRY;
	int		err;

	DEBUG_CMD("%s: Sending data to the card (%d bytes)\n", 
			card->devname, len);
	WAN_MBOX_INIT(mbox);
	do{
		mbox->wan_fr_dlci    = dlci;
		mbox->wan_fr_attr    = attr;
		mbox->wan_data_len  = len;
		mbox->wan_command = FR_WRITE;
		err = card->hw_iface.cmd(card->hw, card->mbox_off+FR_TX_MBOX_OFFS, mbox);
	} while (err && retry-- && !fr_event(card, err, mbox));

	return err;   
}


/* 
 * ============================================================================
 * Send a frame on a selected DLCI.  
 */
static int 
fr_send(sdla_t* card, int dlci, unsigned char attr, int len, void *buf, int snd_encp)
{
	wan_mbox_t*	mbox = &card->wan_mbox;
	bus_size_t	frbuf_off;	
	int 		err;
	wan_smp_flag_t	s;
        
	wan_spin_lock_irq(NULL, &s);
	if (snd_encp){
    		err = fr_write(card, dlci, attr, len+2);
	}else{
		err = fr_write(card, dlci, attr, len);
	}
	if (!err){ 
		fr_tx_buf_ctl_t frbuf;
		unsigned long data_addr = 0l;

		frbuf_off = *(bus_size_t*)mbox->wan_data;
		card->hw_iface.peek(card->hw, frbuf_off, &frbuf, sizeof(frbuf)); 

		/*Add encapsulation header if snd_encp flag set to 1 !!! */
		if (snd_encp){
			unsigned short ip_over_fr_header = FR_IP_HEADER;
			card->hw_iface.poke(card->hw, frbuf.offset, &ip_over_fr_header, 2);
			data_addr += 2;
		}
		card->hw_iface.poke(card->hw, frbuf.offset, buf, len);
		frbuf.flag = 0x01;
		card->hw_iface.poke(card->hw, frbuf_off, &frbuf, sizeof(frbuf));
	}

	wan_spin_unlock_irq(NULL, &s);
	return err;
}

/*============================================================================
 * Read TE1/56K Front end registers
 */
static unsigned char 
read_front_end_reg (void* card1, unsigned short reg)
{
	sdla_t* 	card = (sdla_t*)card1;
	wan_mbox_t* 	mbox = &card->wan_mbox;
	char*		data = mbox->wan_data;
	int 		err = 0x0;

	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = sizeof(FRONT_END_REG_STRUCT);
	mbox->wan_command = READ_FRONT_END_REGISTER;
	((FRONT_END_REG_STRUCT*)data)->register_number = reg;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		fr_event(card, err, mbox);
	}
	return (((FRONT_END_REG_STRUCT*)data)->register_value);
}

/*============================================================================
 * Write to TE1/56K Front end registers  
 */
static unsigned char 
write_front_end_reg (void* card_id, unsigned short reg, unsigned char value)
{
	sdla_t*		card = (sdla_t*)card_id;
	wan_mbox_t*	mbox = &card->wan_mbox;
	char*		data = mbox->wan_data;
	int 		err = 0x00;
	
	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = sizeof(FRONT_END_REG_STRUCT);
	mbox->wan_command = WRITE_FRONT_END_REGISTER;
	((FRONT_END_REG_STRUCT*)data)->register_number = reg;
	((FRONT_END_REG_STRUCT*)data)->register_value = value;
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
        	fr_event(card, err, mbox);
	}

	return err;
}

/*============================================================================
 * TE1 T1/E1 enable hardware interrupt  
 */
static void fr_enable_timer(void* card_id)
{
	sdla_t*		card = (sdla_t*)card_id;
	
	card->u.f.timer_int_enabled |= TMR_INT_ENABLED_TE;
	card->hw_iface.set_bit(card->hw, card->intr_perm_off, FR_INTR_TIMER);
	return;
}

/*============================================================================
 * TE1_56K Set adapter type  
 */
static int set_adapter_config (sdla_t* card)
{
	wan_mbox_t*			mbox = &card->wan_mbox;
	ADAPTER_CONFIGURATION_STRUCT*	cfg = (ADAPTER_CONFIGURATION_STRUCT*)mbox->wan_data;
	int 				err = 0x00;

	card->hw_iface.getcfg(card->hw, SDLA_ADAPTERTYPE, &cfg->adapter_type);
	DEBUG_CMD("%s: Set adapter type (%x type)\n", 
			card->devname,
			cfg->adapter_type);

	WAN_MBOX_INIT(mbox);
	mbox->wan_data_len = sizeof(ADAPTER_CONFIGURATION_STRUCT);
	mbox->wan_command = SET_ADAPTER_CONFIGURATION;
	cfg->adapter_config = 0x00; 
	cfg->operating_frequency = 0x00; 
	err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	if (err != WAN_CMD_OK){
		fr_event(card, err, mbox);
	}

	return (err);
}

/****** Firmware Asynchronous Event Handlers ********************************/

/*
 * ============================================================================
 * Main asyncronous event/error handler.
 *	This routine is called whenever firmware command returns non-zero
 *	return code.
 *
 * Return zero if previous command has to be cancelled.
 */
static int fr_event (sdla_t *card, int event, wan_mbox_t* mbox)
{
	struct wan_dev_le	*devle;
	netdevice_t		*ifp;

	DEBUG_CMD("%s: Asyncronous event/error (0x%x)\n",
			 card->devname, event); 
	switch (event){ 
	case FRRES_MODEM_FAILURE:
        	return fr_modem_failure(card, mbox);

	case FRRES_CHANNEL_DOWN:
	  	DEBUG_EVENT("%s: FR Event: Channel Down!\n",card->devname);
		/* Remove all routes from associated DLCI's */
		WAN_LIST_FOREACH(devle, &card->wandev.dev_head, dev_link){
			fr_channel_t	*chan = wan_netif_priv(devle->dev);
			
        		ifp = WAN_DEVLE2DEV(devle);
			if (!ifp) continue;
			if (chan->route_flag == ROUTE_ADDED){ 
        		        chan->route_flag = REMOVE_ROUTE;
				chan->timer_int_enabled |= TMR_INT_ENABLED_PROCESS_ROUTE;
				ifp->if_timer = 1;
			}

			if (chan->inarp == INARP_CONFIGURED){ 
				chan->inarp = INARP_REQUEST;
			}
		}
		wanpipe_set_state(card, WAN_DISCONNECTED);
		WAN_DEBUG_START(card);
		break;

	case FRRES_CHANNEL_UP:
	{
		int num_requests = 0;
	
		DEBUG_EVENT("%s: FR Event: Channel Up!\n",card->devname);
		/* Remove all routes from associated DLCI's */
       		WAN_LIST_FOREACH(devle, &card->wandev.dev_head, dev_link){
			fr_channel_t *chan = wan_netif_priv(devle->dev);

			ifp = WAN_DEVLE2DEV(devle);
			if (!ifp) continue;
			if( chan->inarp == INARP_REQUEST){
				num_requests++;
				chan->inarp_tick = SYSTEM_TICKS;
			}
		}

		/* Allow timer interrupts */
		if (num_requests){
			card->hw_iface.set_bit(card->hw, card->intr_perm_off, FR_INTR_TIMER);
		}
		wanpipe_set_state(card, WAN_CONNECTED);
		break;
	}

	case FRRES_DLCI_CHANGE:
		return fr_dlci_change(card, mbox);

	case FRRES_DLCI_MISMATCH:
		DEBUG_EVENT("%s: FR Event: DLCI list mismatch!\n", 
				card->devname);
		break;

	case WAN_CMD_TIMEOUT:
	{
		unsigned char str[9];

		DEBUG_EVENT("%s: command 0x%02x timed out!\n", 
				card->devname, mbox->wan_command);
		card->hw_iface.peek(card->hw, card->intr_type_off + 0x18, str, 8); 
		str[8] = '\0';
		DEBUG_EVENT("%s: ID Bytes = %s\n",card->devname, str);
		return 1;
	}

	case FRRES_DLCI_INACTIVE:
	case FRRES_CIR_OVERFLOW:
	case FRRES_BUFFER_OVERFLOW:
		return 1;
 
	default:
		DEBUG_EVENT("%s: command 0x%02x returned 0x%02x!\n",
				card->devname, mbox->wan_command, event);
		return 1;
	} 

	return 0;
}

/* 
 * ============================================================================
 * Handle modem error.
 *
 * Return zero if previous command has to be cancelled.
 */
static int fr_modem_failure (sdla_t *card, wan_mbox_t* mbox)
{

	if (IS_56K_CARD(card)){
		FRONT_END_STATUS_STRUCT FE_status; 
		card->hw_iface.peek(card->hw, 0xF020, &FE_status, sizeof(FE_status));
		card->fe.fe_param.k56_param.RR8_reg_56k = 
			FE_status.FE_U.stat_56k.RR8_56k;	
		card->fe.fe_param.k56_param.RRA_reg_56k = 
			FE_status.FE_U.stat_56k.RRA_56k;	
		card->fe.fe_param.k56_param.RRC_reg_56k = 
			FE_status.FE_U.stat_56k.RRC_56k;	
		card->wandev.fe_iface.read_alarm(&card->fe, 0);
		FE_status.opp_flag = 0x01;
		card->hw_iface.poke(card->hw, 0xF020, &FE_status, sizeof(FE_status));
	}else if (IS_TE1_CARD(card)){
		FRONT_END_STATUS_STRUCT FE_status; 
		card->hw_iface.peek(card->hw, 0xF020, &FE_status, sizeof(FE_status));
		/* TE1 T1/E1 interrupt */
		card->wandev.fe_iface.isr(&card->fe); 

		FE_status.opp_flag = 0x01;
		card->hw_iface.poke(card->hw, 0xF020, &FE_status, sizeof(FE_status));
	}else{
	    	DEBUG_EVENT("%s: physical link down! (modem error 0x%02x)\n",
				card->devname, mbox->wan_data[0]);
		if (mbox->wan_data[0] == 0x28){
			card->fe.fe_status = FE_CONNECTED;
		}else{
			card->fe.fe_status = FE_DISCONNECTED;
		}
	}
	
	switch (mbox->wan_command){
	case FR_WRITE:
	case FR_READ:
		return 1;
	}
	return 0;
}

/* 
 * ============================================================================
 * Handle DLCI status change.
 *
 * Return zero if previous command has to be cancelled.
 */
static int fr_dlci_change (sdla_t *card, wan_mbox_t* mbox)
{
	struct wan_dev_le	*devle;
	fr_channel_t	*chan = NULL;
	netdevice_t	*ifp = NULL;
	int 		cnt = mbox->wan_data_len / sizeof(dlci_status_t);
	dlci_status_t 	status;
	int 		i = 1;
	unsigned short 	dlci;
	unsigned char	state;
	int 		num_requests = 0;

	memcpy(&status, mbox->wan_data, sizeof(dlci_status_t));
	for (i = 1; cnt; --cnt, i++){
		dlci= status.dlci;
		state = status.state;
		ifp = find_channel(card, dlci);
	
		if (ifp == NULL){
        		DEBUG_EVENT("%s: Warning: Unconfigured DLCI %d reported as %s, %s!\n", 
					card->devname, dlci,
					(state & FR_DLCI_DELETED) ? "Deleted" : "Included",
					(state & FR_DLCI_ACTIVE) ? "Active" : "Inactive");

		}else{
			chan = ifp->if_softc;
			if ((state & ~FR_DLCI_NEW) == FR_LINK_INOPER){ 
				DEBUG_EVENT("%s: DLCI %u reported as Included, Inactive%s!\n", 
						card->devname, dlci,
						(state & FR_DLCI_NEW) ? ", New" : "");

			if (chan->start){ 
				set_chan_state(ifp, WAN_DISCONNECTED); 
				WAN_DEBUG_START(card);
			}
		}
	
		if (state & FR_DLCI_DELETED){ 
			DEBUG_EVENT("%s: DLCI %u reported as Deleted!\n", 
						card->devname, dlci);
			if (chan->start){
				if (chan->route_flag == ROUTE_ADDED){ 
					chan->route_flag = REMOVE_ROUTE;
					chan->timer_int_enabled |= TMR_INT_ENABLED_PROCESS_ROUTE;
					ifp->if_timer = 1;
				}

				if (chan->inarp == INARP_CONFIGURED){ 
					chan->inarp = INARP_REQUEST;
				}

				set_chan_state(ifp, WAN_DISCONNECTED);
				WAN_DEBUG_START(card);
			}
		}else 
			if (state & FR_DLCI_ACTIVE){ 
				DEBUG_EVENT("%s: DLCI %u reported as Included, Active%s!\n", 
					card->devname, dlci,
					(state & FR_DLCI_NEW) ? ", New" : "");
				/* This flag is used for configuring specific 
				**  DLCI(s) when they become active.
				*/ 
				chan->dlci_configured = DLCI_CONFIG_PENDING;

				if (chan->start){
					set_chan_state(ifp, WAN_CONNECTED);
				}
			}
		}
		
		memcpy(&status, mbox->wan_data + i * sizeof(dlci_status_t), sizeof(dlci_status_t));
#if 0
#ifdef __OpenBSD__
	READ_BUF(card->mbox + offsetof(wan_mbox_t, wan_data) + i * sizeof(dlci_status_t),
    					&status, sizeof(dlci_status_t));
#endif
#endif
	}
	
       	WAN_LIST_FOREACH(devle, &card->wandev.dev_head, dev_link){
		ifp = WAN_DEVLE2DEV(devle);
		if (!ifp) continue;
		chan = wan_netif_priv(ifp);
		if (chan->dlci_configured == DLCI_CONFIG_PENDING){
			if (!fr_init_dlci(card, chan)){
				num_requests++;
			}
		}
	}

	/* Allow timer interrupts */
	if (num_requests){
		card->u.f.timer_int_enabled = TMR_INT_ENABLED_INARP;
		card->hw_iface.set_bit(card->hw, card->intr_perm_off, FR_INTR_TIMER);
	}
	return 0;
}

/*
 * =============================================================================
 * fr_init_dlci() - 
 *
 */
static int fr_init_dlci(sdla_t *card, fr_channel_t *chan)
{
	fr_dlc_conf_t cfg;
	    
	bzero(&cfg, sizeof(cfg));

	if ( chan->cir_status == CIR_DISABLED){ 
		cfg.cir_fwd = cfg.cir_bwd  = 16;
		cfg.bc_fwd = cfg.bc_bwd = 16;
		cfg.conf_flags = 0x0001;
	}else{ 
		if (chan->cir_status == CIR_ENABLED){ 
			cfg.cir_fwd = cfg.cir_bwd = chan->cir;
			cfg.bc_fwd  = cfg.bc_bwd  = chan->bc;
			cfg.be_fwd  = cfg.be_bwd  = chan->be;
			cfg.conf_flags = 0x0000;
		}
	}

	if (fr_dlci_configure( card, &cfg , chan->dlci)){
		DEBUG_EVENT("%s: DLCI Configure failed for %d\n",
					card->devname, chan->dlci);
		return 1;
	}

	chan->dlci_configured = DLCI_CONFIGURED;

	/* Allow timer interrupts */
	if (chan->inarp == INARP_REQUEST && card->wandev.state == WAN_CONNECTED){ 
		chan->inarp_tick = SYSTEM_TICKS;
	}	

	/* Read the interface byte mapping into the channel 
	 * structure.
	 */
	read_DLCI_IB_mapping(card, chan);
	return 0;
 }

/******* Miscellaneous ******************************************************/

/* 
 * ===========================================================================
 * Update channel state. 
 */
static int update_chan_state (struct ifnet* ifp)
{
	fr_channel_t* 	chan = ifp->if_softc;
	sdla_t*		card = chan->common.card;
	wan_mbox_t*	mbox = &card->wan_mbox;
	int 		retry = MAX_CMD_RETRY;
	int 		err;

	WAN_MBOX_INIT(mbox);
	do{
		mbox->wan_command = FR_LIST_ACTIVE_DLCI;
		mbox->wan_data_len = 0;
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	} while (err && retry-- && !fr_event(card, err, mbox));
	if (!err){ 
		unsigned short 	offset = card->mbox_off + 
					offsetof(wan_mbox_t, wan_data);
		unsigned 	dlci;
		int 		cnt = mbox->wan_data_len / sizeof(short);

		for (; cnt; --cnt, ++offset){ 
			card->hw_iface.peek(card->hw,
					    offset,
					    &dlci,
					    sizeof(dlci));
			if (dlci == chan->dlci){ 
 	        		set_chan_state(ifp, WAN_CONNECTED);
			        break;
			}
		}
	}

	return err;
}


/* 
 * ============================================================================
 * Set channel state.
 */
static void set_chan_state (struct ifnet* ifp, int state)
{
	fr_channel_t*	chan = ifp->if_softc;
	sdla_t*		card = chan->common.card;
	unsigned long flags;

	flags = read_eflags();
	disable_intr();
	if (chan->common.state != state){ 
		switch (state){ 
		case WAN_CONNECTED:
			DEBUG_EVENT("%s:%s: Set DLCI %d connected\n",
					card->devname, chan->name, chan->dlci);
			fr_up(ifp);
			break;

		case WAN_CONNECTING:
			DEBUG_EVENT("%s:%s: Set DLCI %d connecting...\n",
					card->devname, chan->name, chan->dlci);
			ifp->if_flags |= IFF_RUNNING;
			break;

		case WAN_DISCONNECTED:
			DEBUG_EVENT("%s:%s: Set DLCI %d disconnected!\n",
					card->devname, chan->name, chan->dlci);
			fr_down(ifp);
   	    		break;
		}
		chan->common.state = state;
	}

	chan->state_tick = SYSTEM_TICKS;
	write_eflags(flags);
}

/* 
 *============================================================================
 * Find network device by its channel number.
 */
static struct ifnet* find_channel (sdla_t* card, unsigned dlci)
{
    if (dlci > HIGHEST_VALID_DLCI){
	return NULL;
    }
    return(card->u.f.dlci_to_dev_map[dlci]);
}

/* 
 * ============================================================================
 * Check to see if a frame can be sent. If no transmit buffers available,
 * enable transmit interrupts.
 *
 * Return:	1 - Tx buffer(s) available
 *		0 - no buffers available
 */
static int is_tx_ready (sdla_t* card, fr_channel_t* chan)
{
	unsigned char sb;

	if (card->type == SDLA_S508){
		card->hw_iface.isa_read_1(card->hw, 0x00, &sb);
    		if (sb & 0x02){ 
			return 1;
		}
		return 0;
	}
	return 1;
}

/* 
 * ============================================================================
 * Convert decimal string to unsigned integer.
 * If len != 0 then only 'len' characters of the string are converted.
 */
static unsigned int dec_to_uint (unsigned char* str, int len)
{
	unsigned val;

	if (!len){ 
		len = strlen(str);
	}

	for (val = 0; len && is_digit(*str); ++str, --len){
		val = (val * 10) + (*str - (unsigned)'0');
	}

	return val;
}


/* 
 * =============================================================================
 * Store a UDP management packet for later processing.
 */
static int 
store_udp_mgmt_pkt(int udp_type, char udp_pkt_src, sdla_t* card, 
				   struct mbuf *m, int dlci)
{
	int udp_pkt_stored = 0;
	int data_length = m->m_pkthdr.len;

	if (!card->u.f.udp_pkt_len && (data_length <= MAX_LGTH_UDP_MGNT_PKT)){ 
		card->u.f.udp_pkt_len = data_length;
	        card->u.f.udp_type = udp_type;
		card->u.f.udp_pkt_src = udp_pkt_src;
       		 card->u.f.udp_dlci = dlci;
		m_copydata(m, 0, data_length, (caddr_t)&card->u.f.udp_pkt_data);
		card->u.f.timer_int_enabled = TMR_INT_ENABLED_UDP;
		udp_pkt_stored = 1;
	}else{
		DEBUG_EVENT("%s: UDP packet not stored for DLCI %d\n", 
					card->devname, dlci);
	}

	wan_skb_free(m);	/*m_freem(m);*/
	
	return (udp_pkt_stored);
}

/*
 * Execute Frame Relay udp commands
 */
static int fr_udp_ioctl(struct ifnet* ifp, caddr_t data)
{
    	fr_channel_t* 		chan = ifp->if_softc;
    	sdla_t* 	  	card = chan->common.card;
	struct ifreq*		ifr = (struct ifreq*)data;
	wan_udp_hdr_t		udp_hdr;
	int 			error = 0;

	if (copyin((void*)ifr->ifr_data, (void*)&udp_hdr, sizeof(wan_udp_hdr_t)) < 0){
		DEBUG_EVENT("%s: Failed to copy data from user to kernel space\n",
				card->devname);
		return -EFAULT;
	}

	error = process_udp_cmd(ifp, &udp_hdr);	

	if (copyout((void*)&udp_hdr, (void*)ifr->ifr_data, sizeof(wan_udp_hdr_t)) < 0){
		DEBUG_EVENT("%s: Failed to copy data from kernel to user space\n",
				card->devname);
		return -EFAULT;
	}
	return 0;
}

static int process_udp_cmd(struct ifnet* ifp, wan_udp_hdr_t* udp_hdr)
{
    	fr_channel_t* 		chan = ifp->if_softc;
    	sdla_t* 	  	card = chan->common.card;
    	wan_mbox_t* 		mbox = &card->wan_mbox;
    	fr_trc_el_t* 		ptr_trc_el = NULL;
    	fr_trc_el_t 		trc_el;
    	fpipemon_trc_t* 	fpipemon_trc = NULL;
    	unsigned short 		num_trc_els;
    	unsigned char 		frames;
    	unsigned short 		buffer_length = 0;
    	int 			c_retry = MAX_CMD_RETRY;
	int			err = 0;

	switch(udp_hdr->wan_udphdr_command){ 
	case FPIPE_ENABLE_TRACING:
		if (!card->TracingEnabled){ 
			WAN_MBOX_INIT(mbox);
			do{
				mbox->wan_command = FR_SET_TRACE_CONFIG;
				mbox->wan_data_len = 1;
				mbox->wan_fr_dlci = 0x00;
				mbox->wan_data[0] = udp_hdr->wan_udphdr_data[0]|RESET_TRC;
				err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
			} while (err && c_retry-- && !fr_event(card, err, mbox));

			if (err){ 
				card->TracingEnabled = 0;
				/* set the return code */
				udp_hdr->wan_udphdr_return_code = mbox->wan_return_code;
				udp_hdr->wan_udphdr_data_len = 0;
				break;
			}

			card->hw_iface.peek(card->hw,
				    NO_TRC_ELEMENTS_OFF,
				    &num_trc_els,
				    2);
			card->hw_iface.peek(card->hw,
				    BASE_TRC_ELEMENTS_OFF,
				    &card->u.f.trc_el_base,
				    4);
			card->u.f.curr_trc_el = card->u.f.trc_el_base;
                	card->u.f.trc_el_last = 
				(unsigned char*)card->u.f.curr_trc_el + 
				((num_trc_els - 1) * sizeof(fr_trc_el_t));
   
			/* Calculate the maximum trace data area in */
			/* the UDP packet */
	        	card->u.f.trc_bfr_space=(MAX_LGTH_UDP_MGNT_PKT -
				/*sizeof(fr_encap_hdr_t) -*/
				sizeof(struct ip) -
				sizeof(struct udphdr) -
				sizeof(wan_mgmt_t) -
				sizeof(wan_cmd_t));

			/* set return code */
			udp_hdr->wan_udphdr_return_code = 0;
		}else{ 
			/* set return code to line trace already enabled */
			udp_hdr->wan_udphdr_return_code = 1;
        	}
		udp_hdr->wan_udphdr_data_len = 0;
		card->TracingEnabled = 1;
		break;

	case FPIPE_DISABLE_TRACING:
		if(card->TracingEnabled){ 
			WAN_MBOX_INIT(mbox);
			do{
				mbox->wan_command = FR_SET_TRACE_CONFIG;
				mbox->wan_data_len = 1;
				mbox->wan_fr_dlci = 0x00;
				mbox->wan_data[0] = ~ACTIVATE_TRC;
				err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
			} while (err && c_retry-- && !fr_event(card, err, mbox));
		}

		/* set return code */
		udp_hdr->wan_udphdr_return_code = 0;
		udp_hdr->wan_udphdr_data_len = 0;
		card->TracingEnabled = 0;
		break;

	case FPIPE_GET_TRACE_INFO:
		/* Line trace cannot be performed on the 502 */
        	if (!card->TracingEnabled){ 
			/* set return code */
                	udp_hdr->wan_udphdr_return_code = 1;
	    		udp_hdr->wan_udphdr_data_len = 0;
                	break;
		}

		(void *)ptr_trc_el = card->u.f.curr_trc_el;

        	buffer_length = 0;
		udp_hdr->wan_udphdr_data[0x00] = 0x00;

        	for(frames = 0; frames < MAX_FRMS_TRACED; frames ++){ 
			card->hw_iface.peek(card->hw,
				    (unsigned long)ptr_trc_el, 
				    (void *)&trc_el.flag,
				    sizeof(fr_trc_el_t));
			if (trc_el.flag == 0x00) 
				break;
			if ((card->u.f.trc_bfr_space - buffer_length) < 
						    sizeof(fpipemon_trc_hdr_t)){ 
				udp_hdr->wan_udphdr_data[0x00] |= MORE_TRC_DATA;
                		break;
			}

			fpipemon_trc = 
				(fpipemon_trc_t *)&udp_hdr->wan_udphdr_data[buffer_length]; 
			fpipemon_trc->fpipemon_trc_hdr.status = trc_el.attr;
               		fpipemon_trc->fpipemon_trc_hdr.tmstamp = trc_el.tmstamp;
			fpipemon_trc->fpipemon_trc_hdr.length = trc_el.length;

			if (!trc_el.offset || !trc_el.length){ 
				fpipemon_trc->fpipemon_trc_hdr.data_passed = 0x00;
			}else{ 
				if ((trc_el.length + sizeof(fpipemon_trc_hdr_t) + 1) >
		        			(card->u.f.trc_bfr_space - buffer_length)){
					fpipemon_trc->fpipemon_trc_hdr.data_passed = 0x00;
					udp_hdr->wan_udphdr_data[0x00] |= MORE_TRC_DATA;
				}else{
					fpipemon_trc->fpipemon_trc_hdr.data_passed = 0x01;
                			card->hw_iface.peek(card->hw, 
							trc_el.offset, 
							fpipemon_trc->data,
				 			trc_el.length);
				}	
			}
			trc_el.flag = 0x00;
                	card->hw_iface.poke_byte(card->hw, 
					 (unsigned long)ptr_trc_el, 
					 0x00);
			ptr_trc_el ++;
			if ((void *)ptr_trc_el > card->u.f.trc_el_last){
				(void*)ptr_trc_el = card->u.f.trc_el_base;
			}

			buffer_length += sizeof(fpipemon_trc_hdr_t);
			if (fpipemon_trc->fpipemon_trc_hdr.data_passed){ 
				buffer_length += trc_el.length;
			}

			if (udp_hdr->wan_udphdr_data[0x00] & MORE_TRC_DATA){ 
				break;
			}
        	}
                      
		if(frames == MAX_FRMS_TRACED){ 
			udp_hdr->wan_udphdr_data[0x00] |= MORE_TRC_DATA;
		}
       
		card->u.f.curr_trc_el = (void *)ptr_trc_el;

        	/* set the total number of frames passed */
        	udp_hdr->wan_udphdr_data[0x00] |= ((frames << 1) & (MAX_FRMS_TRACED << 1));

		/* set the data length and return code */
		udp_hdr->wan_udphdr_data_len = buffer_length;
        	udp_hdr->wan_udphdr_return_code = 0;
        	break;

	case FPIPE_FT1_READ_STATUS:
	    card->hw_iface.peek(card->hw, 
				0xF020, 
				&udp_hdr->wan_udphdr_data[0x00], 
				2);
	    udp_hdr->wan_udphdr_data_len = 2;
	    udp_hdr->wan_udphdr_return_code = 0;
	    break;

	case FPIPE_FLUSH_DRIVER_STATS:
	    init_chan_statistics(chan);
	    init_global_statistics(card);
	    udp_hdr->wan_udphdr_return_code = 0;
	    break;

	case FPIPE_ROUTER_UP_TIME:
  	    wan_getcurrenttime(&chan->router_up_time, NULL);
	    chan->router_up_time -= chan->router_start_time;
	    *(unsigned long *)&udp_hdr->wan_udphdr_data[0] = chan->router_up_time;	
	    udp_hdr->wan_udphdr_data_len = sizeof(unsigned long);
	    udp_hdr->wan_udphdr_return_code = 0;
	    break;

	case FR_FT1_STATUS_CTRL:
	    if (udp_hdr->wan_udphdr_data[0] == 1){ 
		if (rCount++ != 0){
	            udp_hdr->wan_udphdr_return_code = 0;
	    	    udp_hdr->wan_udphdr_data_len = 1;
		    break;
		} 
	    }
           
	    /* Disable FT1 MONITOR STATUS */
            if (udp_hdr->wan_udphdr_data[0] == 0){ 
	        if ( --rCount != 0){ 
		    udp_hdr->wan_udphdr_return_code = 0;
	    	    udp_hdr->wan_udphdr_data_len = 1;
		    break;
	        } 
	    }  
	    goto udp_dflt_cmd;
	    break;

	case FPIPE_DRIVER_STAT_IFSEND:
	    bcopy(&chan->drvstats_if_send.if_send_entry, 
		  &udp_hdr->wan_udphdr_data[0], 
		   sizeof(if_send_stat_t));
	    udp_hdr->wan_udphdr_data_len = sizeof(if_send_stat_t);
	    udp_hdr->wan_udphdr_return_code = 0;
	    break;

	case FPIPE_DRIVER_STAT_INTR:
	    bcopy(&card->statistics.isr_entry, &udp_hdr->wan_udphdr_data[0],
		   sizeof(global_stats_t));
	    bcopy(&chan->drvstats_rx_intr.rx_intr_no_socket,
	          &udp_hdr->wan_udphdr_data[sizeof(global_stats_t)],
		   sizeof(rx_intr_stat_t));
	    udp_hdr->wan_udphdr_data_len = sizeof(global_stats_t) + sizeof(rx_intr_stat_t);
	    udp_hdr->wan_udphdr_return_code = 0;
	    break;

	case FPIPE_DRIVER_STAT_GEN:
            bcopy(&chan->drvstats_gen.UDP_PIPE_mgmt_kmalloc_err,
	    	  &udp_hdr->wan_udphdr_data[0], sizeof(pipe_mgmt_stat_t)); 
	    bcopy(&card->statistics,
	          &udp_hdr->wan_udphdr_data[sizeof(pipe_mgmt_stat_t)],
		  sizeof(global_stats_t));
	    udp_hdr->wan_udphdr_return_code = 0;
	    udp_hdr->wan_udphdr_data_len = sizeof(global_stats_t) +
					sizeof(rx_intr_stat_t);
	    break;

		case WAN_FE_SET_LB_MODE:
	case WAN_GET_MEDIA_TYPE:
	case WAN_FE_GET_STAT:
	case WAN_FE_GET_CFG:
 	case WAN_FE_FLUSH_PMON:
    	    if (IS_TE1_CARD(card)){
		card->wandev.fe_iface.process_udp(
					&card->fe,
					&udp_hdr->wan_cmd, 
					&udp_hdr->wan_udphdr_data[0]);	
	    }else if (IS_56K_CARD(card)){
		card->wandev.fe_iface.process_udp(
					&card->fe,
					&udp_hdr->wan_cmd, 
					&udp_hdr->wan_udphdr_data[0]);	
	    }else{
		if (udp_hdr->wan_udphdr_command == WAN_GET_MEDIA_TYPE){
	    		udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
    	    		udp_hdr->wan_udphdr_data_len = sizeof(unsigned char); 
		}else{
	    		udp_hdr->wan_udphdr_return_code = WAN_UDP_INVALID_CMD;
		}	
	    }
	    break;
#if 0
	case WAN_FE_SET_LB_MODE:
	    /* Activate/Deactivate Line Loopback modes */
	    if (IS_TE1_CARD(card)){
		err = sdla_set_te1_lb_modes(card, 
					    udp_hdr->wan_udphdr_data[0], 
					    udp_hdr->wan_udphdr_data[1]);
	    	udp_hdr->wan_udphdr_return_code = 
				(!err) ? WAN_CMD_OK : WAN_UDP_FAILED_CMD;
	    }else{
	    	udp_hdr->wan_udphdr_return_code = WAN_UDP_INVALID_CMD;
	    }
	    udp_hdr->wan_udphdr_data_len = 0x00;
	    break;

  	case WAN_GET_MEDIA_TYPE:
	    udp_hdr->wan_udphdr_data[0] = 
		(IS_T1_CARD(card) ? WANOPT_MEDIA_T1 :
		 IS_E1_CARD(card) ? WANOPT_MEDIA_E1 :
		 IS_56K_CARD(card) ? WANOPT_MEDIA_56K : 
					WANOPT_MEDIA_NONE);
	    udp_hdr->wan_udphdr_return_code = 0;
    	    udp_hdr->wan_udphdr_data_len = sizeof(unsigned char); 
	    break;

	case WAN_FE_GET_STAT:
 	    /* TE1_56K Read T1/E1/56K alarms */
	    if (IS_TE1_CARD(card)){	
	  	*(unsigned long *)&udp_hdr->wan_udphdr_data[0] = sdla_te_alarm(card, 0); 
		/* TE1 Update T1/E1 perfomance counters */
    		sdla_te_pmon(card);
	        bcopy(&card->wandev.te_pmon,
	        	  &udp_hdr->wan_udphdr_data[sizeof(unsigned long)],
			  sizeof(pmc_pmon_t));
		udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
	    	udp_hdr->wan_udphdr_data_len = 
			sizeof(unsigned long) + sizeof(pmc_pmon_t);

	    }else if (IS_56K_CARD(card)) {
			/* 56K Update CSU/DSU alarms */
		card->wandev.k56_alarm = sdla_56k_alarm(card, 1); 
		*(unsigned long *)&udp_hdr->wan_udphdr_data[0] = card->wandev.k56_alarm;
		udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
	    	udp_hdr->wan_udphdr_data_len = sizeof(unsigned long);
	    }else{
	    	udp_hdr->wan_udphdr_return_code = WAN_UDP_INVALID_CMD;
	    	udp_hdr->wan_udphdr_data_len = 0;
	    }
	    break;

	case WAN_FE_FLUSH_PMON:
 	    /* TE1 Flush T1/E1 pmon counters */
	    if (IS_TE1_CARD(card)){	
		sdla_flush_te1_pmon(card);
	        udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
	    }else{
	    	udp_hdr->wan_udphdr_return_code = WAN_UDP_INVALID_CMD;
	    	udp_hdr->wan_udphdr_data_len = 0;
	    }
	    break;

	case WAN_FE_GET_CFG:
	    /* Read T1/E1 configuration */
	    if (IS_TE1_CARD(card)){	
            	bcopy(&card->wandev.te_cfg,
        	      &udp_hdr->wan_udphdr_data[0],
		      sizeof(sdla_te_cfg_t));
	    	udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
    		udp_hdr->wan_udphdr_data_len = sizeof(sdla_te_cfg_t);
	    }else{
	    	udp_hdr->wan_udphdr_return_code = WAN_UDP_INVALID_CMD;
	    	udp_hdr->wan_udphdr_data_len = 0;
	    }
	    break;
#endif

 	case WAN_GET_PROTOCOL:
	    udp_hdr->wan_udphdr_data[0] = WANCONFIG_FR;
	    udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
    	    udp_hdr->wan_udphdr_data_len = 1;
	    break;

 	case WAN_GET_PLATFORM:
	    udp_hdr->wan_udphdr_data[0] = WAN_FREEBSD_PLATFORM;
	    udp_hdr->wan_udphdr_return_code = WAN_CMD_OK;
    	    udp_hdr->wan_udphdr_data_len = 1;
	    break;

      default:
udp_dflt_cmd:
		WAN_MBOX_INIT(mbox);
		do{
			mbox->wan_command 	= udp_hdr->wan_udphdr_command;
			mbox->wan_data_len	= udp_hdr->wan_udphdr_data_len;
			mbox->wan_fr_dlci 	= udp_hdr->wan_udphdr_fr_dlci;
			mbox->wan_fr_attr 	= udp_hdr->wan_udphdr_fr_attr;
			mbox->wan_fr_rxlost1 	= udp_hdr->wan_udphdr_fr_rxlost1;
			mbox->wan_fr_rxlost2 	= udp_hdr->wan_udphdr_fr_rxlost2;
	        	if (mbox->wan_data_len){ 
				memcpy(mbox->wan_data,
					(char *)udp_hdr->wan_udphdr_data,
					mbox->wan_data_len);
			}
			err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
		} while (err && c_retry-- && !fr_event(card, err, mbox));

		if (!err){
			chan->drvstats_gen.UDP_PIPE_mgmt_adptr_cmnd_OK ++;
		}else{
			chan->drvstats_gen.UDP_PIPE_mgmt_adptr_cmnd_timeout ++;
		}

		/* copy the result back to our buffer */
		udp_hdr->wan_udphdr_data_len   = mbox->wan_data_len;
		udp_hdr->wan_udphdr_return_code= mbox->wan_return_code;
		udp_hdr->wan_udphdr_fr_dlci    = mbox->wan_fr_dlci;
		udp_hdr->wan_udphdr_fr_attr    = mbox->wan_fr_attr;
		udp_hdr->wan_udphdr_fr_rxlost1 = mbox->wan_fr_rxlost1;
  		udp_hdr->wan_udphdr_fr_rxlost2 = mbox->wan_fr_rxlost2;

		if (mbox->wan_data_len){ 
			memcpy(&udp_hdr->wan_udphdr_data[0],
				mbox->wan_data,
				mbox->wan_data_len);
		}
	} 

	return err;
}


/* 
 * =============================================================================
 * Process UDP call of type FPIPE8ND
 */
static int process_udp_mgmt_pkt(sdla_t* card)
{
	struct ifnet* 	ifp = NULL;
	fr_channel_t* 	chan = NULL;
	struct mbuf*	m = NULL; 
	wan_udp_pkt_t*	udp_pkt = NULL;
	int 		udp_mgmt_req_valid = 1;
	char 		udp_pkt_src = card->u.f.udp_pkt_src; 
	int 		dlci = card->u.f.udp_dlci;
	unsigned char*	data_addr = NULL;
	unsigned int 	data_length = 0;
	int 		err;

	/* Find network interface for this packet */
	ifp = find_channel(card, dlci);
	chan = ifp->if_softc;
	udp_pkt = (wan_udp_pkt_t *)&card->u.f.udp_pkt_data;
	
	/*
	 * If the UDP packet is from the network, we are going to have to transmit
	 * a response. Before doing so, we must check to see that we are not 
	 * currently transmitting a frame (in 'if_send()') and that we are not 
	 * already in a 'delayed transmit' state.
	 */
	if (udp_pkt_src == UDP_PKT_FRM_NETWORK){
		if ((ifp->if_flags & IFF_OACTIVE) || (card->u.f.tx_interrupts_pending))
			return 0; 
	}

	if (udp_pkt_src == UDP_PKT_FRM_NETWORK){ 
    		switch(udp_pkt->wan_udp_command){ 
		case FR_READ_MODEM_STATUS:
		case FR_READ_STATUS:
		case FPIPE_ROUTER_UP_TIME:
		case FR_READ_ERROR_STATS:
		case FPIPE_DRIVER_STAT_GEN:
		case FR_READ_STATISTICS:
		case FR_READ_ADD_DLC_STATS:
		case FR_READ_CONFIG:
		case FR_READ_CODE_VERSION:
		case WAN_GET_MEDIA_TYPE:
		case WAN_FE_GET_STAT:
			udp_mgmt_req_valid = 1;
			break;
	
		default:
			udp_mgmt_req_valid = 0;
			break;
		}
	}

	udp_pkt->wan_udp_opp_flag = 0;
	if (!udp_mgmt_req_valid){ 
		/* set length to 0 */
		udp_pkt->wan_udp_data_len = 0;
		chan->drvstats_gen.UDP_PIPE_mgmt_direction_err ++;
		/* set return code */
		udp_pkt->wan_udp_return_code = 0xCD;
	}else{ 
	
		err = process_udp_cmd(ifp, &udp_pkt->wan_udp_hdr);

	}
   
	/* Fill UDP TTL */
	udp_pkt->wan_ip_ttl = card->wandev.ttl;

	data_length = reply_udp(&card->u.f.udp_pkt_data, udp_pkt->wan_udp_data_len);

	if (udp_pkt_src == UDP_PKT_FRM_NETWORK){ 
		err = fr_send(card, dlci, 0, data_length, (unsigned char*)&card->u.f.udp_pkt_data, 1);
		if (err){ 
			chan->drvstats_gen.UDP_PIPE_mgmt_adptr_send_passed ++;
		}else{
			chan->drvstats_gen.UDP_PIPE_mgmt_adptr_send_failed ++;
		}
	}else{ 
#if defined(__NetBSD__) || defined(__OpenBSD__) || (__FreeBSD_version < 501000)
		wan_smp_flag_t	s;
#endif

		data_addr = (unsigned char*)&card->u.f.udp_pkt_data;

		m = sdla_get(ifp, data_addr, data_length);
		if (m == NULL){ 
			chan->drvstats_gen.UDP_PIPE_mgmt_no_socket ++;
			DEBUG_EVENT("%s: %s: UDP mgmt cmnd, no socket buffers available!\n",
					card->devname, chan->name);
			/* i guess nothing can be done ... the packet is gone */
			return ENOMEM;
		}
		wan_bpf_report(ifp, m, 1);
		/* 
		 * Ok, at this point we've got an mbuf chain pointed by m which contains
		 * a UDP management reply - now we just have to pass it up the stack -
		 * and that's it !!!
		 * Note that we neither report those packets to BPF nore update any 
		 * stat. Those packets are meant to be invisible ;-)
		 * Also we run inside splimp already - do there's no need to call it 
		 * again.
		 */
#if (__FreeBSD_version >= 501000)
		netisr_queue(NETISR_IP, m);
#else
		wan_spin_lock_irq(NULL, &s);
		schednetisr(NETISR_IP);
		if (IF_QFULL(&ipintrq)){ 
			chan->drvstats_gen.UDP_PIPE_mgmt_not_passed_to_stack ++;
			/* oh, no - IP queue is full - well - screw it. 
			 * We can't do much here. 
			 */ 
			DEBUG_RX("%s: %s: IP queue is full, drop packet!\n", 
					card->devname, chan->name);
			IF_DROP(&ipintrq);
			wan_skb_free(m);	/*m_freem(m);*/
			wan_spin_unlock_irq(NULL, &s);
			return ENOMEM;
		}
  
		/* ok enqueue the packet */
		IF_ENQUEUE(&ipintrq, m);
		wan_spin_unlock_irq(NULL, &s);
#endif
		DEBUG_UDP("%s: %s: UDP packet passed back on IP's queue\n", 
				card->devname, chan->name);
		chan->drvstats_gen.UDP_PIPE_mgmt_passed_to_stack ++;
	}

	card->u.f.udp_pkt_len = 0;

	return 1;
}

/* 
 * ============================================================================
 * send_inarp_request()- sends out an InARP request packet on specified
 *			    interface
 */
static void 
send_inarp_request(sdla_t *card, struct ifnet *ifp)
{
    struct ifaddr*	ifa = WAN_TAILQ_FIRST(ifp);
    struct sockaddr_in*	saddr = NULL;	/* the address */
    arphdr_fr_t        	arphdr;
    fr_channel_t*	chan = ifp->if_softc;

    /* set up the pointer to the local address */
    DEBUG_TX("%s: %s: Sending InARP request...\n", 
				card->devname, chan->name);
    if ((ifa = WAN_TAILQ_NEXT(ifa)) == NULL)
    {
	DEBUG_EVENT("%s: The local address didn't find!\n",
				chan->name);
	return;
    }
    saddr = (struct sockaddr_in *)ifa->ifa_addr;

    /* InARP request */
    arphdr.ar_hrd = 0x0F00;	            /* Frame Relay HW type */
    arphdr.ar_pro = 0x0008;	       	    /* IP Protocol         */
    arphdr.ar_hln = 2;		 	    /* HW addr length      */
    arphdr.ar_pln = 4;		 	    /* IP addr length      */
    arphdr.ar_op = htons(0x08);	 	    /* InARP Request       */
    arphdr.ar_sha = 0;		 	    /* src HW DLCI - Doesn't matter */
    arphdr.ar_sip = saddr->sin_addr.s_addr; /* Local Address       */
    arphdr.ar_tha = 0;			    /* dst HW DLCI - Doesn't matter */
    arphdr.ar_tip = 0;			    /* Remote Address -- what we want */
  
    /* send it out ... */
    send_inarp_packet(card, &arphdr, ifp);

    /* done */
    return;
}

/* 
 * ============================================================================
 * send_inarp_packet() - sends previously formatted InARP packet on 
 *			    specified interface 
 */
static void     
send_inarp_packet(sdla_t *card, arphdr_fr_t * arphdr, struct ifnet *ifp)
{
    fr_channel_t* 	chan = ifp->if_softc;
    arphdr_1490_t*	ArpPacket = NULL;
    unsigned short	data_length;	 /* data length */
    int err = 0;

    data_length = sizeof(arphdr_1490_t) + sizeof(arphdr_fr_t);

    /* allocate space for the packet */
    if ((ArpPacket = wan_malloc(data_length)) == NULL){ 
        /* oops ... well we'll retry later */
	goto send_failed;
    }

    /* prepare the header */
    ArpPacket->control = 0x03;
    ArpPacket->pad     = 0x00;
    ArpPacket->NLPID   = 0x80;
    ArpPacket->OUI[0]  = 0;
    ArpPacket->OUI[1]  = 0;
    ArpPacket->OUI[2]  = 0;
    ArpPacket->PID     = 0x0608;

    /* copy the packet */
    bcopy(arphdr, ArpPacket + 1, sizeof(arphdr_fr_t));

    DEBUG_TX("%s:%s: Sending InARP request on DLCI %d.\n", 
			card->devname, chan->name, chan->dlci);
    err = fr_send(card, chan->dlci, 0, 
		  sizeof(arphdr_1490_t) + sizeof(arphdr_fr_t),
		  (void *)ArpPacket, 0);
    if (err)
        goto send_failed;

    /* see if it was InARP request - if so - flip the flag indicating that */
    /* we are awaiting InARP reply ...					 */
    if (arphdr->ar_op == htons(0x08)){ 
	DEBUG_TX("%s:%s: Sent InARP request ...\n", 
				card->devname, chan->name); 
	chan->fr_inarp_reply_expected = 1;
	chan->fr_inarp_reply_timer    = INARP_REQUEST_TIMEOUT;
	/* schedule watchdog */
	ifp->if_timer = 1; 		/* call it in a second */
    }else{
        DEBUG_EVENT("%s:%s: Sent InARP reply ...\n", 
			card->devname, chan->name); 
    }

    /* free the packet and exit */
    if (ArpPacket)
    	wan_free(ArpPacket);

    /* done */
    return;

send_failed:

    /* ok, if the send failed for some reason - what we gonna do is to */
    /* copy the InARP packet into dlci_scp->fr_inarp_pkt and schedule  */
    /* watchdog which will try to resend it ...			     */
    DEBUG_EVENT("%s:%s: Failed to send InARP packet ..scheduling watchdog.\n", 
				card->devname, chan->name);

    /* copy the packet and set the flag for the watchdog */
    bcopy(arphdr, &chan->fr_inarp_pkt, sizeof(arphdr_fr_t));
    chan->fr_resend_inarp = 1;

    /* schedule watchdog */
    ifp->if_timer = 1; 		/* call it in a second */

    /*  free the pkt if needed and exit */
    if (ArpPacket){ 
	wan_free(ArpPacket);
    }

    /* done */
    return;
}

/* 
 * =============================================================================
 * Check packet for ARP Type
 */

int is_arp(void *buf)
{
    arphdr_1490_t *arphdr = (arphdr_1490_t *)buf;

    if (arphdr->pad   == 0x00  &&
        arphdr->NLPID == 0x80  &&
        arphdr->PID   == 0x0608) 
	return 1;
    else return 0;
}

/* 
 *=============================================================================
 * Process ARP Packet Type
 */

int process_ARP(arphdr_1490_t *ArpPacket, sdla_t *card, struct ifnet* ifp)
{
    struct ifaddr*	ifa = NULL;
    struct sockaddr_in* local_addr = NULL, *dest_addr = NULL, *net_mask = NULL;
    arphdr_fr_t*	arphdr = (arphdr_fr_t *)(ArpPacket + 1); /* Skip header */
    fr_rx_buf_ctl_t 	frbuf;
    fr_channel_t*	chan = ifp->if_softc;

    /* get pointers to the IP addresses and masks */
    if ((ifa = WAN_TAILQ_FIRST(ifp)) == NULL){ 
	return EFAULT;
    }
    if ((ifa = WAN_TAILQ_NEXT(ifa)) == NULL){
	return EFAULT;
    }
    local_addr = (struct sockaddr_in *)ifa->ifa_addr;
    dest_addr  = (struct sockaddr_in *)ifa->ifa_dstaddr;
    net_mask   = (struct sockaddr_in *)ifa->ifa_netmask;

    switch (ntohs(arphdr->ar_op)){ 
    case 0x08:  /* Inverse ARP request  -- Send Reply, add route. */
	/* Check for valid Address */
        DEBUG_EVENT("%s:%s: RX PtP addr %s -InArp Req\n", 
			card->devname, chan->name, in_ntoa(arphdr->ar_sip));

        /* ok, let's see if the sender's IP is on the same subnet as ours 
         * if not - we don't reply and write a log message  	      
         */
        if ((local_addr->sin_addr.s_addr & net_mask->sin_addr.s_addr) != 
	    (arphdr->ar_sip & net_mask->sin_addr.s_addr)) 
        {
	    DEBUG_EVENT("%s:%s: Invalid PtP address.  InARP ignored.\n", 
					card->devname, chan->name);
            return -1;
        }

        if (local_addr->sin_addr.s_addr == arphdr->ar_sip){
	    DEBUG_EVENT("%s:%s: Local addr = PtP addr.  InARP ignored.\n", 
					card->devname, chan->name);
            return -1;
        } 
        arphdr->ar_op = htons(0x09);    /* InARP Reply */

        /* Set addresses */
        arphdr->ar_tip = arphdr->ar_sip;
        arphdr->ar_sip = local_addr->sin_addr.s_addr;
        card->hw_iface.peek(card->hw, card->rxmb_off, &frbuf, sizeof(fr_rx_buf_ctl_t));
        fr_send(card, frbuf.dlci, 0, frbuf.length, (void *)ArpPacket, 0);

        /* Modify Point-to-Point Address */
        chan->dst_addr = arphdr->ar_tip;

        /* Add Route Flag */
        /* The route will be added in the polling routine so
         * that it is not interrupt context. 
         */
        chan->route_flag = ADD_ROUTE;
        /* now if the interface has InARP enabled - we would like to update 
         * routing and interface information to reflect new dest. ip   	
         */
	chan->timer_int_enabled |= TMR_INT_ENABLED_PROCESS_ROUTE;
        ifp->if_timer = 1;

        /* done */
	break;

    case 0x09:  /*  Inverse ARP reply */
	/* Check for valid Address */
        DEBUG_EVENT("%s:%s: RX PtP addr %s -InArp Reply\n", 
				card->devname, chan->name, in_ntoa(arphdr->ar_sip));

	/* reset the flag ... */
	chan->fr_inarp_reply_expected = 0;

        /* ok, let's see if the sender's IP is on the same subnet as ours
	 * if not - we don't reply and write a log message  	      
         */
        if ((net_mask->sin_addr.s_addr & arphdr->ar_sip) != 
	    (net_mask->sin_addr.s_addr & local_addr->sin_addr.s_addr)) 
	{
	    DEBUG_EVENT("%s:%s: Invalid PtP address. InARP ignored.\n", 
					card->devname, chan->name);
            return -1;
        }

        if (local_addr->sin_addr.s_addr == arphdr->ar_sip){ 
	    DEBUG_EVENT("%s:%s: Local addr = PtP addr.  InARP ignored.\n", 
					card->devname, chan->name);
            return -1;
        }

        /* Modify Point-to-Point Address */
 	dest_addr->sin_addr.s_addr = arphdr->ar_sip;

        /* Add Route Flag */
        /* The route will be added in the polling routine so
         * that it is not interrupt context. 
	 */
	chan->route_flag = ADD_ROUTE;
        chan->inarp = INARP_CONFIGURED;
        /* The route will be added in the polling routine so
         * that it is not interrupt context. 
         */
	chan->timer_int_enabled |= TMR_INT_ENABLED_PROCESS_ROUTE;
	ifp->if_timer = 1;

        /* Modify Point-to-Point Address */
	chan->dst_addr = arphdr->ar_sip;
        break;

    default:  /* ARP's and RARP's -- Shouldn't happen. */
        break;
    }

    return 0;
}

#ifdef FR_OLD_INTR_TEST
/* 
 * ============================================================================
 * Perform the Interrupt Test by running the READ_CODE_VERSION command 
 * MAX_INTR_TEST_COUNTER times.
 */
static int fr_intr_test( sdla_t* card )
{
	int err,i;

	err = fr_set_intr_mode(card, FR_INTR_READY, card->wandev.mtu, 0 );
	
	if (err == WAN_CMD_OK){ 
		wan_mbox_t*	mbox = &card->wan_mbox;
		WAN_MBOX_INIT(mbox);
		for ( i = 0; i < MAX_INTR_TEST_COUNTER; i++ ){ 
			/* Run command READ_CODE_VERSION */
			mbox->wan_data_len  = 0;
			mbox->wan_command = FR_READ_CODE_VERSION;
			err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
			if (err != WAN_CMD_OK){ 
				!fr_event(card, err, mbox);
			}
		}
	}else{ 
		return err;
	}

	err = fr_set_intr_mode(card, 0, card->wandev.mtu, 0);

	if (err != WAN_CMD_OK){ 
		return err;
	}

	return 0;
}
#endif

/* 
 * ============================================================================
 * Determine what type of UDP call it is. DRVSTATS or FPIPE8ND ?
 */
static int udp_pkt_type(sdla_t* card, struct mbuf *m)
{
    wan_udp_pkt_t *udp_pkt = (wan_udp_pkt_t *)mtod(m, wan_udp_pkt_t *);

    if (wan_skb_len(m) < sizeof(wan_udp_pkt_t)){
    	return UDP_INVALID_TYPE;
    }
    if ((udp_pkt->wan_ip_p == UDPMGMT_UDP_PROTOCOL) &&
        (udp_pkt->wan_ip_v == 0x4) &&
        (udp_pkt->wan_udp_request_reply == UDPMGMT_REQUEST)) 
    {
        if ((udp_pkt->wan_udp_dport == ntohs(card->wandev.udp_port)) &&
            (!strncmp(udp_pkt->wan_udp_signature, FR_UDP_SIGNATURE, FR_UDP_SIGNATURE_LEN) ||
             !strncmp(udp_pkt->wan_udp_signature, GLOBAL_UDP_SIGNATURE, GLOBAL_UDP_SIGNATURE_LEN))){
	    return UDP_FPIPE_TYPE;
	}
        if ((udp_pkt->wan_udp_dport == ntohs(card->wandev.udp_port)) &&
	    !strncmp(udp_pkt->wan_udp_signature, UDP_DRVRSTATS_SIGNATURE, 8)){
	    return UDP_DRVSTATS_TYPE;
	}
        if (!strncmp(udp_pkt->wan_udp_signature, GLOBAL_UDP_SIGNATURE, GLOBAL_UDP_SIGNATURE_LEN) && 
		udp_pkt->wan_udp_command == WAN_GET_PROTOCOL){
	    return UDP_FPIPE_TYPE;
	}
    }

    return UDP_INVALID_TYPE;
}

/*
 * =============================================================================
 * Initializes the Statistics values in the fr_channel structure.
 */
void init_chan_statistics( fr_channel_t* chan)
{
    bzero(&chan->drvstats_if_send.if_send_entry, 
				    sizeof(drvstats_if_send_t));
    bzero(&chan->drvstats_rx_intr.rx_intr_no_socket, 
				    sizeof(drvstats_rx_intr_t));
    bzero(&chan->drvstats_gen.UDP_PIPE_mgmt_kmalloc_err, 
			            sizeof(pipe_mgmt_stat_t));
}
	
/* 
 * =============================================================================
 * Initializes the Statistics values in the Sdla_t structure.
 */
void init_global_statistics( sdla_t* card )
{
	/* Intialize global statistics for a card */
	bzero(&card->statistics.isr_entry, sizeof(global_stats_t));
}

/* 
 * =============================================================================
 * 
 */
static void read_DLCI_IB_mapping( sdla_t* card, fr_channel_t* chan )
{
	wan_mbox_t*		mbox = &card->wan_mbox;
	dlci_IB_mapping_t 	result; 
	int			offset = 0, i = 0;
	int 			retry = MAX_CMD_RETRY;	
	int 			err, counter, found = 0;	

	WAN_MBOX_INIT(mbox);
	do{
		mbox->wan_command = FR_READ_DLCI_IB_MAPPING;
		mbox->wan_data_len = 0;	
		err = card->hw_iface.cmd(card->hw, card->mbox_off, mbox);
	} while (err && retry-- && !fr_event(card, err, mbox));

	if (mbox->wan_return_code != 0){
		DEBUG_EVENT("%s:%s: Read DLCI IB Mapping failed\n",
				card->devname, chan->name); 
	}
	counter = mbox->wan_data_len / sizeof(dlci_IB_mapping_t);
	memcpy(&result, mbox->wan_data, sizeof(dlci_IB_mapping_t));
	offset = card->mbox_off + offsetof(wan_mbox_t, wan_data);
	for (i = 1; counter; --counter, i++){ 
		if (result.dlci == chan->dlci){ 
			chan->IB_addr = result.addr_value;
			chan->dlci_int_interface_off = chan->IB_addr; 
			found = 1;
			break;
		} 
		card->hw_iface.peek(
				card->hw, 
				offset + i * sizeof(dlci_IB_mapping_t),
				&result, 
				sizeof(dlci_IB_mapping_t));
	}

	if (!found){
		DEBUG_EVENT("%s: DLCI %d not found by IB MAPPING cmd\n", 
						card->devname, chan->dlci);
	}
}

static void fr_handle_front_end_state (void* card_id)
{
	sdla_t*		card = (sdla_t*)card_id;

	WAN_ASSERT1(card == NULL);
	/* FIXME: Add handling front end status changes */

	return;
}

/*
 ******************************************************************************
 *			wan_snmp_data()	
 *
 * Description: Save snmp request and parameters in private structure, enable
 *		TIMER interrupt, put current process in sleep.
 * Arguments:   
 * Returns:
 ******************************************************************************
 */
#define   FRDLCMIIFINDEX        3
#define   FRDLCMISTATE          4
#define   FRDLCMIADDRESS        5
#define   FRDLCMIADDRESSLEN     6
#define   FRDLCMIPOLLINGINTERVAL  7
#define   FRDLCMIFULLENQUIRYINTERVAL  8
#define   FRDLCMIERRORTHRESHOLD  9
#define   FRDLCMIMONITOREDEVENTS  10
#define   FRDLCMIMAXSUPPORTEDVCS  11
#define   FRDLCMIMULTICAST      12
#define   FRDLCMISTATUS         13
#define   FRDLCMIROWSTATUS      14
#define   FRCIRCUITIFINDEX      17
#define   FRCIRCUITDLCI         18
#define   FRCIRCUITSTATE        19
#define   FRCIRCUITRECEIVEDFECNS  20
#define   FRCIRCUITRECEIVEDBECNS  21
#define   FRCIRCUITSENTFRAMES   22
#define   FRCIRCUITSENTOCTETS   23
#define   FRCIRCUITRECEIVEDFRAMES  24
#define   FRCIRCUITRECEIVEDOCTETS  25
#define   FRCIRCUITCREATIONTIME  26
#define   FRCIRCUITLASTTIMECHANGE  27
#define   FRCIRCUITCOMMITTEDBURST  28
#define   FRCIRCUITEXCESSBURST  29
#define   FRCIRCUITTHROUGHPUT   30
#define   FRCIRCUITMULTICAST    31
#define   FRCIRCUITTYPE         32
#define   FRCIRCUITDISCARDS     33
#define   FRCIRCUITRECEIVEDDES  34
#define   FRCIRCUITSENTDES      35
#define   FRCIRCUITLOGICALIFINDEX  36
#define   FRCIRCUITROWSTATUS    37
#define   FRERRIFINDEX          40
#define   FRERRTYPE             41
#define   FRERRDATA             42
#define   FRERRTIME             43
#define   FRERRFAULTS           44
#define   FRERRFAULTTIME        45
#define   FRTRAPSTATE           46
#define   FRTRAPMAXRATE         47

static int fr_snmp_data(sdla_t* card, struct ifnet* ifp, void* data)
{
    	fr_channel_t* 	chan = ifp->if_softc;
	wanpipe_snmp_t*	snmp = NULL;
	
	/* Update device statistics */
	if (card->wandev.update) {
		int rslt = 0;
		card->u.f.update_dlci = chan;
		rslt = card->wandev.update(&card->wandev);
		if(rslt) {
			return (rslt) ? (-EBUSY) : (-EINVAL);
		}
	}

	snmp = (wanpipe_snmp_t*)data;

	switch(snmp->snmp_magic){
	/***** Data Link Connection Management Interface ****/
	case FRDLCMIIFINDEX: 
		break;

	case FRDLCMISTATE:
		 snmp->snmp_val = 
		 (card->wandev.signalling == WANOPT_FR_ANSI) ? SNMP_FR_ANSIT1617D :
		 (card->wandev.signalling == WANOPT_FR_Q933) ? SNMP_FR_ITUT933A :
		 (card->wandev.signalling == WANOPT_FR_LMI) ? SNMP_FR_LMIREV:
		       						SNMP_FR_NOLMICONF;
		break;

	case FRDLCMIADDRESS:
		snmp->snmp_val = SNMP_FR_Q922;
		break;

	case FRDLCMIADDRESSLEN:
		snmp->snmp_val = SNMP_FR_4BYTE_ADDR;
		break;

	case FRDLCMIPOLLINGINTERVAL:
		snmp->snmp_val = card->u.f.t391;
		break;

	case FRDLCMIFULLENQUIRYINTERVAL:
		snmp->snmp_val = card->u.f.n391;
		break;

	case FRDLCMIERRORTHRESHOLD:
		snmp->snmp_val = card->u.f.n392;
		break;

	case FRDLCMIMONITOREDEVENTS:
		snmp->snmp_val = card->u.f.n393;
		break;

	case FRDLCMIMAXSUPPORTEDVCS:
		snmp->snmp_val = HIGHEST_VALID_DLCI;
		break;

	case FRDLCMIMULTICAST:
		snmp->snmp_val = SNMP_FR_NONBROADCAST;
		break;

	case FRDLCMISTATUS: /* FIXME */
		snmp->snmp_val = SNMP_FR_RUNNING;
		break;

	case FRDLCMIROWSTATUS: /* FIXME */
		snmp->snmp_val = 0;
		break;

	/****************** Circuit Table *******************/
	case FRCIRCUITIFINDEX:
		break;

	case FRCIRCUITDLCI:
		snmp->snmp_val = chan->dlci;
		break;

	case FRCIRCUITSTATE: /* FIXME */
		snmp->snmp_val = SNMP_FR_ACTIVE;
		break;

	case FRCIRCUITRECEIVEDFECNS:
		snmp->snmp_val = chan->rx_FECN;
		break;

	case FRCIRCUITRECEIVEDBECNS:
		snmp->snmp_val = chan->rx_BECN;
		break;

	case FRCIRCUITSENTFRAMES:
		snmp->snmp_val = ifp->if_opackets;
		break;

	case FRCIRCUITSENTOCTETS:
		snmp->snmp_val = ifp->if_obytes;
		break;

	case FRCIRCUITRECEIVEDFRAMES:
		snmp->snmp_val = ifp->if_ipackets;
		break;

	case FRCIRCUITRECEIVEDOCTETS:
		snmp->snmp_val = ifp->if_ibytes;
		break;

	case FRCIRCUITCREATIONTIME:
		snmp->snmp_val = chan->router_start_time;
		break;

	case FRCIRCUITLASTTIMECHANGE:
		snmp->snmp_val = chan->router_last_change;
		break;

	case FRCIRCUITCOMMITTEDBURST:
		snmp->snmp_val = (unsigned long)chan->bc;
		break;

	case FRCIRCUITEXCESSBURST:
		snmp->snmp_val = (unsigned long)chan->be;
		break;

	case FRCIRCUITTHROUGHPUT: /* FIXME */
		snmp->snmp_val = (unsigned long)0;
		break;

	case FRCIRCUITMULTICAST:
		snmp->snmp_val = (chan->mc) ? SNMP_FR_ONEWAY : SNMP_FR_UNICAST;
		break;

	case FRCIRCUITTYPE:
		snmp->snmp_val = chan->dlci_type;
		break;

	case FRCIRCUITDISCARDS:
		snmp->snmp_val = ifp->if_ierrors;
		break;

	case FRCIRCUITRECEIVEDDES:
		snmp->snmp_val = chan->rx_DE_set;
		break;

	case FRCIRCUITSENTDES:
		snmp->snmp_val = chan->tx_DE_set;
		break;

	case FRCIRCUITLOGICALIFINDEX: /* FIXME */
		snmp->snmp_val = 0;
		break;

	case FRCIRCUITROWSTATUS: /* FIXME */
		snmp->snmp_val = 0;
		break;

	/****************** Errort Table  *******************/
	case FRERRIFINDEX: 
		break;

	case FRERRTYPE: /* FIXME */
		snmp->snmp_val = chan->err_type;
		break;

	case FRERRDATA: /* FIXME */
		memcpy((void*)snmp->snmp_data, 
			chan->err_data, 
			strlen(chan->err_data));
		break;

	case FRERRTIME: /* FIXME */
		snmp->snmp_val = chan->err_time;
		break;

	case FRERRFAULTS: /* FIXME */
		snmp->snmp_val = chan->err_faults;
		break;

	case FRERRFAULTTIME: /* FIXME */
		snmp->snmp_val = (unsigned long)0;
		break;

	/************ Frame Relay Trap Control **************/
	case FRTRAPSTATE:
		snmp->snmp_val = chan->trap_state;
		break;

	case FRTRAPMAXRATE:
		snmp->snmp_val = chan->trap_max_rate;
		break;

	default:
            	return -EAFNOSUPPORT;
	}

	return 0;

}

/*
****************************************************************************
**
**
**
*/
static int fr_debugging(sdla_t* card)
{
	if (card->wan_debugging_state == WAN_DEBUGGING_CONT && card->wandev.station == WANOPT_CPE){
		card->wan_debugging_state = WAN_DEBUGGING_PROTOCOL;
		/*
		** If link still down after 100 seconds (only for CPE), then
		** report error. 
		*/
		return 100;
	}
	if (card->wandev.station == WANOPT_CPE){
		if (card->wan_debug_last_msg != WAN_DEBUG_FR_CPE_MSG){
			DEBUG_EVENT("%s: Frame Relay switch not replying.\n",
						card->devname);
			DEBUG_EVENT("%s: Check ANSI/LMI/Q.933 setting with Frame Relay provider.\n",
						card->devname);
		}
		card->wan_debug_last_msg = WAN_DEBUG_FR_CPE_MSG;
	}else{
		if (card->wan_debug_last_msg != WAN_DEBUG_FR_NODE_MSG){
			DEBUG_EVENT("%s: No Signalling frames received from Frame Relay CPE!\n",
						card->devname);
		}
		card->wan_debug_last_msg = WAN_DEBUG_FR_NODE_MSG;
	} 
	return 0;
}

static unsigned long fr_crc_frames(sdla_t* card)
{
	fr_comm_stat_t	err_stats;
	if (!fr_get_err_stats(card, &err_stats)){
		return err_stats.rx_bad_crc;
	}
	return 0;
}

static unsigned long fr_abort_frames(sdla_t * card)
{
	fr_comm_stat_t	err_stats;
	if (!fr_get_err_stats(card, &err_stats)){
		return err_stats.rx_aborts;
	}
	return 0;
}

static unsigned long fr_tx_underun_frames(sdla_t* card)
{
	fr_comm_stat_t	err_stats;
	if (!fr_get_err_stats(card, &err_stats)){
		return err_stats.tx_missed_undr;
	}

	return 0;
}

/*
 * ******************************************************************
 * Proc FS function 
 */
static int wan_fr_get_info(void* pcard, struct seq_file *m, int *stop_cnt)
{
	sdla_t	*card = (sdla_t*)pcard;

	if (card->wandev.fe_iface.update_alarm_info){
		m->count = 
			WAN_FECALL(
				&card->wandev, 
				update_alarm_info, 
				(&card->fe, m, stop_cnt)); 
	}
	if (card->wandev.fe_iface.update_pmon_info){
		m->count = 
			WAN_FECALL(
				&card->wandev, 
				update_pmon_info, 
				(&card->fe, m, stop_cnt)); 
	}

	return m->count;
}
/****** End *****************************************************************/
